from abc import ABC
from collections.abc import Mapping

import numpy as np
import torch
from ftag import Flavours
from numpy.lib.recfunctions import unstructured_to_structured as u2s
from torch import Tensor, nn

from salt.models import Dense
from salt.stypes import Tensors
from salt.utils.array_utils import listify
from salt.utils.class_names import CLASS_NAMES
from salt.utils.scalers import RegressionTargetScaler
from salt.utils.tensor_utils import masked_softmax
from salt.utils.union_find import get_node_assignment_jit


class TaskBase(nn.Module, ABC):
    def __init__(
        self,
        name: str,
        input_name: str,
        dense_config: dict,
        loss: nn.Module,
        weight: float = 1.0,
    ):
        """Task head base class.

        Tasks wrap a dense network, a loss, a label, and a weight.

        Parameters
        ----------
        name : str
            Arbitrary name of the task, used for logging and inference.
        input_name : str
            Which type of object is input to the task e.g. jet/track/flow.
        dense_config : dict
            Keyword arguments for [`salt.models.Dense`][salt.models.Dense],
            the dense network producing the task outputs.
        loss : nn.Module
            Loss function applied to the dense network outputs.
        weight : float
            Weight in the overall loss.
        """
        super().__init__()

        self.name = name
        self.input_name = input_name
        self.net = Dense(**dense_config)
        self.loss = loss
        self.weight = weight

    def input_name_mask(self, pad_masks: Mapping):
        return torch.cat(
            [
                torch.ones(m.shape[1], device=m.device) * (1 if (t == self.input_name) else 0)
                for t, m in pad_masks.items()
            ],
        ).bool()


class ClassificationTask(TaskBase):
    def __init__(
        self,
        label: str,
        class_names: list[str] | None = None,
        label_map: Mapping | None = None,
        sample_weight: str | None = None,
        use_class_dict: bool = False,
        **kwargs,
    ):
        """Classification task.

        Parameters
        ----------
        label : str
            Label name for the task
        class_names : list[str] | None, optional
            List of class names, ordered by output index. If not specified attempt to
            automatically determine these from the label name.
        label_map : Mapping | None, optional
            Remap integer labels for training (e.g. 0,4,5 -> 0,1,2).
        sample_weight : str | None, optional
            Name of a per sample weighting to apply in the loss function.
        use_class_dict : bool, optional
            If True, read class weights for the loss from the class_dict file.
        **kwargs
            Keyword arguments for [`salt.models.TaskBase`][salt.models.TaskBase].
        """
        super().__init__(**kwargs)
        self.label = label
        self.class_names = class_names
        self.label_map = label_map
        if self.label_map is not None and self.class_names is None:
            raise ValueError("Specify class names when using label_map.")
        if hasattr(self.loss, "ignore_index"):
            self.loss.ignore_index = -1
        self.sample_weight = sample_weight
        if self.sample_weight is not None:
            assert (
                self.loss.reduction == "none"
            ), "Sample weights only supported for reduction='none'"
        if self.class_names is None:
            self.class_names = CLASS_NAMES[self.label]
        if len(self.class_names) != self.net.output_size:
            raise ValueError(
                f"{self.name}: "
                f"Number of outputs ({self.net.output_size}) does not match "
                f"number of class names ({len(self.class_names)}). Class names: {self.class_names}"
            )
        self.use_class_dict = use_class_dict

    @property
    def output_names(self) -> list[str]:
        assert self.class_names is not None
        pxs = [f"{Flavours[c].px}" if c in Flavours else f"p{c}" for c in self.class_names]
        return [f"{self.model_name}_{px}" for px in pxs]

    def apply_sample_weight(self, loss: Tensor, labels_dict: Mapping) -> Tensor:
        """Apply per sample weights, if specified."""
        if self.sample_weight is None:
            return loss
        return (loss * labels_dict[self.input_name][self.sample_weight]).mean()

    def forward(
        self,
        x: Tensor,
        labels_dict: Mapping,
        pad_masks: Mapping | None = None,
        context: Tensor = None,
    ):
        # get predictions and mask
        if pad_masks is not None:
            input_name_mask = self.input_name_mask(pad_masks)
            preds = self.net(x[:, input_name_mask], context)
            pad_mask = pad_masks[self.input_name]
        else:
            preds = self.net(x, context)
            pad_mask = None

        # get labels and remap them if necessary
        labels = labels_dict[self.input_name][self.label] if labels_dict else None
        if labels is not None and self.label_map is not None:
            mapped_labels = torch.clone(labels)
            for k, v in self.label_map.items():
                mapped_labels[labels == k] = v
            labels = mapped_labels

        # use the mask to remove padded values from the loss (ignore_index=-1 is set by default)
        if pad_mask is not None and labels is not None:
            # mask out dodgey labels
            # TODO: remove when is in the samples
            # https://gitlab.cern.ch/atlas/athena/-/merge_requests/60199
            pad_mask = torch.masked_fill(pad_mask, labels == -2, True)

            # update the labels based on the mask (in case not done already)
            labels = torch.masked_fill(labels, pad_mask, -1)

        loss = None
        if labels is not None:
            if preds.ndim == 3:
                loss = self.loss(preds.permute(0, 2, 1), labels)
            else:
                loss = self.loss(preds, labels)
            loss = self.apply_sample_weight(loss, labels_dict)
            loss *= self.weight

        return preds, loss

    def run_inference(self, preds: Tensor, pad_mask: Tensor | None = None) -> Tensor:
        if pad_mask is None:
            assert preds.ndim == 2
            probs = torch.softmax(preds, dim=-1)
        else:
            assert preds.ndim == 3
            probs = masked_softmax(preds, pad_mask.unsqueeze(-1))
        return probs

    def get_h5(self, preds: Tensor, pad_mask: Tensor | None = None) -> np.ndarray:
        probs = self.run_inference(preds, pad_mask)
        dtype = np.dtype([(n, "f4") for n in self.output_names])
        return u2s(probs.float().cpu().numpy(), dtype)

    def get_onnx(self, preds: Tensor, **kwargs) -> tuple:
        probs = self.run_inference(preds, kwargs.get("pad_mask", None))
        return tuple(output.squeeze() for output in torch.split(probs, 1, -1))


class RegressionTaskBase(TaskBase, ABC):
    def __init__(
        self,
        targets: list[str] | str,
        scaler: RegressionTargetScaler | None = None,
        target_denominators: list[str] | str | None = None,
        norm_params: dict | None = None,
        custom_output_names: list[str] | str | None = None,
        **kwargs,
    ):
        """Base class for regression tasks.

        Parameters
        ----------
        targets : list[str] | str
            Regression target(s).
        scaler : RegressionTargetScaler
            Functional scaler for regression targets
                - cannot be used with other target scaling options.
        target_denominators : list[str] | str | None, optional
            Variables to divide regression target(s) by (i.e. for regressing a ratio).
                - cannot be used with other target scaling options.
        norm_params : dict | None, optional
            Mean and std normalization parameters for each target, used for scaling.
                - cannot be used with other target scaling options.
        custom_output_names : list[str] | str | None, optional
            Name(s) of regression output(s), overwrites the standard "model name + numerator"
        **kwargs
            Keyword arguments for [`salt.models.TaskBase`][salt.models.TaskBase].
        """
        super().__init__(**kwargs)
        self.scaler = scaler
        self.targets = listify(targets)
        self.target_denominators = listify(target_denominators)
        self.custom_output_names = listify(custom_output_names)
        if norm_params:
            norm_params["mean"] = listify(norm_params["mean"])
            norm_params["std"] = listify(norm_params["std"])
        self.norm_params = norm_params

        if [scaler, target_denominators, norm_params].count(None) not in {2, 3}:
            raise ValueError("Can only use a single scaling method")

        if self.scaler:
            for target in self.targets:
                self.scaler.scale(target, torch.Tensor(1))
        if self.target_denominators and len(self.targets) != len(self.target_denominators):
            raise ValueError(
                f"{self.name}: "
                f"Number of targets ({len(self.targets)}) does not match "
                f"number of target denominators ({len(self.target_denominators)})"
            )
        if self.norm_params and len(self.norm_params["mean"]) != len(self.targets):
            raise ValueError(
                f"{self.name}: "
                f"Number of means in norm_params ({len(self.norm_params['mean'])}) does not match "
                f"number of targets ({len(self.targets)})"
            )
        if self.norm_params and len(self.norm_params["std"]) != len(self.targets):
            raise ValueError(
                f"{self.name}: "
                f"Number of stds in norm_params ({len(self.norm_params['std'])}) does not match "
                f"number of targets ({len(self.targets)})"
            )

    def nan_loss(self, preds: Tensor, targets: Tensor, **kwargs) -> Tensor:
        """Calculates the loss function, and excludes any NaNs.

        If Nans are included in the targets, then the loss should be instansiated
        with the `reduction="none"` option, and this function will take the mean
        excluding any nans.
        """
        invalid = torch.isnan(targets)
        preds = torch.where(invalid, torch.zeros_like(preds), preds)
        targets = torch.where(invalid, torch.zeros_like(targets), targets)

        if "var" in kwargs:
            kwargs["var"] = torch.where(invalid, torch.zeros_like(kwargs["var"]), kwargs["var"])

        loss = self.loss(preds, targets, **kwargs)

        if len(loss.shape) == 0:
            if torch.isnan(loss):
                raise ValueError(
                    "Regression loss is NaN. This may be due to NaN targets,"
                    " check configs/nan_regression.yaml for options to deal with this."
                )
            return loss

        nanmean = torch.nanmean(loss)
        if torch.isnan(nanmean):
            raise ValueError("NanRegression is NaN. This means all model predictions are NaN")
        return nanmean

    def get_targets(self, targets_dict: Mapping):
        targets = None
        if targets_dict:
            targets = torch.stack(
                [targets_dict[self.input_name][target] for target in self.targets], dim=1
            )

        if targets is not None:
            if self.scaler is not None:
                for i in range(len(self.targets)):
                    targets[:, i] = self.scaler.scale(self.targets[i], targets[:, i])
            if self.target_denominators is not None:
                for i in range(len(self.targets)):
                    targets[:, i] = torch.div(
                        targets[:, i], targets_dict[self.input_name][self.target_denominators[i]]
                    )
            if self.norm_params is not None:
                for i in range(len(self.norm_params["mean"])):
                    targets[:, i] = (targets[:, i] - self.norm_params["mean"][i]) / (
                        self.norm_params["std"][i]
                    )

            # We stack targets dict always over the first dimension to allow consistency
            # when scaling, but for queries we want the regression target to be in the final
            # dimension. This allows us to keep the same code for both global and query scaling
            if len(targets.shape) == 3:
                targets = targets.transpose(1, 2)
        return targets


class RegressionTask(RegressionTaskBase):
    def __init__(self, scaler=None, **kwargs):
        """Regression task.

        Parameters
        ----------
        scaler
            dummy text
        **kwargs
            Keyword arguments for
            [`salt.models.RegressionTaskBase`][salt.models.RegressionTaskBase].
        """
        super().__init__(**kwargs)
        if self.net.output_size != len(self.targets):
            raise ValueError(
                f"{self.name}: "
                f"Number of outputs ({self.net.output_size}) does not match "
                f"number of targets ({len(self.targets)})"
            )
        self.scaler = scaler

    @property
    def output_names(self) -> list[str]:
        if self.custom_output_names is not None:
            assert len(self.custom_output_names) == len(self.targets)
            return [f"{self.model_name}_{x}" for x in self.custom_output_names]
        return [f"{self.model_name}_{x}" for x in self.targets]

    def forward(
        self,
        x: Tensor,
        targets_dict: Mapping,
        pad_masks: Mapping | None = None,
        context: Tensor = None,
    ):
        if (x.ndim != 2 or pad_masks is not None) and self.input_name != "objects":
            raise NotImplementedError(
                "Regression tasks are currently only supported for global object "
                "and object query level predictions."
            )
        if pad_masks is not None:
            pass
        preds = self.net(x, context)
        targets = self.get_targets(targets_dict)

        loss = None
        if targets is not None:
            loss = self.nan_loss(preds, targets) * self.weight

        return preds, loss

    def run_inference(self, preds: Tensor, labels: Tensors | None = None) -> Tensor:
        preds = preds.float()
        if self.target_denominators is not None and labels is not None:
            for i in range(len(self.targets)):
                preds[:, i] *= labels[self.input_name][self.target_denominators[i]]
        elif self.norm_params is not None:
            for i in range(len(self.norm_params["mean"])):
                preds[:, i] *= self.norm_params["std"][i]
                preds[:, i] += self.norm_params["mean"][i]
        elif self.scaler is not None:
            for i in range(len(self.targets)):
                preds[:, :, i] = self.scaler.inverse(self.targets[i], preds[:, :, i])
        return preds

    def get_h5(self, preds: Tensor, labels: Tensors) -> np.ndarray:
        preds = self.run_inference(preds, labels)
        dtype = np.dtype([(x, "f4") for x in self.output_names])
        return u2s(preds.float().cpu().numpy(), dtype)

    def get_onnx(self, preds: Tensor, **kwargs) -> tuple:
        means = self.run_inference(preds, kwargs.get("labels", None))
        return tuple(output.squeeze() for output in torch.split(means, 1, -1))


class GaussianRegressionTask(RegressionTaskBase):
    def __init__(self, **kwargs):
        """Regression task that outputs a mean and variance for each target.
        The loss function is the negative log likelihood of a Gaussian distribution.

        Parameters
        ----------
        **kwargs
            Keyword arguments for
            [`salt.models.RegressionTaskBase`][salt.models.RegressionTaskBase].
        """
        super().__init__(**kwargs)
        if self.net.output_size != 2 * len(self.targets):
            raise ValueError(
                f"{self.name}: "
                f"Number of targets ({len(self.targets)}) is not twice the "
                f"number of outputs ({self.net.output_size})"
            )

    @property
    def output_names(self) -> list[str]:
        outputs = [*self.targets, *[f"{x}_stddev" for x in self.targets]]
        return [f"{self.model_name}_{x}" for x in outputs]

    def forward(
        self,
        x: Tensor,
        targets_dict: Mapping,
        pad_masks: Mapping | None = None,
        context: Tensor = None,
    ):
        if x.ndim != 2 or pad_masks is not None:
            raise NotImplementedError(
                "Regression tasks are currently only supported for global object level predictions."
            )

        preds = self.net(x, context)
        targets = self.get_targets(targets_dict)

        # split outputs into means and sigmas
        means, variances = preds.tensor_split(2, -1)
        variances = nn.functional.softplus(variances)  # ensure positiveness of variance

        loss = None
        if targets is not None:
            loss = self.nan_loss(means, targets, var=variances) * self.weight

        return preds, loss

    def run_inference(self, preds: Tensor, labels: Tensors | None = None) -> tuple[Tensor, Tensor]:
        if self.target_denominators is not None and labels is not None:
            for i in range(len(self.targets)):
                preds[:, i] *= labels[self.input_name][self.target_denominators[i]]
                preds[:, i + 1] *= labels[self.input_name][self.target_denominators[i]]
        elif self.norm_params is not None:
            for i in range(len(self.norm_params["mean"])):
                preds[:, i] *= self.norm_params["std"][i]
                preds[:, i] += self.norm_params["mean"][i]
                # return stddev as sqrt(var)
                preds[:, i + 1] = (
                    torch.sqrt(nn.functional.softplus(preds[:, i + 1])) * self.norm_params["std"][i]
                )
        else:
            raise ValueError("Inference for Gaussian regression requires scaling parameters.")
        means, stds = preds.tensor_split(2, -1)
        return means, stds

    def get_h5(self, preds: Tensor, labels: Tensors | None = None) -> np.ndarray:
        means, stds = self.run_inference(preds, labels)
        mean_dtype = np.dtype([(f"{self.name}_{t}", "f4") for t in self.targets])
        stds_dtype = np.dtype([(f"{self.name}_{t}_stddev", "f4") for t in self.targets])
        means = u2s(means.float().cpu().numpy(), mean_dtype)
        stds = u2s(stds.float().cpu().numpy(), stds_dtype)
        return means, stds

    def get_onnx(self, preds: Tensor, **kwargs) -> tuple:
        # This might need to be fixed
        # to run inference correctly with denominators
        means, stds = self.run_inference(preds, kwargs.get("labels", None))
        return means.squeeze(), stds.squeeze()


class VertexingTask(TaskBase):
    def __init__(self, label: str, **kwargs):
        """Edge classification task for vertexing.

        Parameters
        ----------
        label : str
            Label name for the target object IDs.
        **kwargs
            Keyword arguments for [`salt.models.TaskBase`][salt.models.TaskBase].
        """
        super().__init__(**kwargs)
        self.label = label

    def forward(
        self,
        x: Tensor,
        labels_dict: Mapping,
        pad_masks: Tensor = None,
        context: Tensor = None,
    ):
        if pad_masks is not None:
            input_name_mask = self.input_name_mask(pad_masks)
            mask = pad_masks[self.input_name]
            x = x[:, input_name_mask]
        else:
            mask = None
        b, n, d = x.shape
        ex_size = (b, n, n, d)
        t_mask = torch.ones(b, n, device=x.device) if mask is None else ~mask
        t_mask = torch.cat(
            [t_mask, torch.zeros(b, 1, device=x.device)], dim=1
        )  # pad t_mask for onnx compatibility
        adjmat = t_mask.unsqueeze(-1) * t_mask.unsqueeze(-2)
        adjmat = (
            adjmat.bool() & ~torch.eye(n + 1, n + 1, device=adjmat.device).repeat(b, 1, 1).bool()
        )

        # Deal with context
        context_matrix = None
        if context is not None:
            context_d = context.shape[-1]
            context = context.unsqueeze(1).expand(b, n, context_d)
            context_matrix = torch.zeros(
                (adjmat.sum(), 2 * context_d), device=x.device, dtype=x.dtype
            )
            context_matrix = context.unsqueeze(-2).expand((b, n, n, context_d))[adjmat[:, :-1, :-1]]

        # Create the track-track matrix as a compressed tensor
        tt_matrix = torch.zeros((adjmat.sum(), d * 2), device=x.device, dtype=x.dtype)
        tt_matrix[:, :d] = x.unsqueeze(-2).expand(ex_size)[adjmat[:, :-1, :-1]]
        tt_matrix[:, d:] = x.unsqueeze(-3).expand(ex_size)[adjmat[:, :-1, :-1]]
        pred = self.net(tt_matrix, context_matrix)
        loss = None
        if labels_dict:
            loss = self.calculate_loss(pred, labels_dict, adjmat=adjmat[:, :-1, :-1])

        return pred, loss

    def calculate_loss(self, pred, labels_dict, adjmat):
        labels = labels_dict[self.input_name][self.label]

        match_matrix = labels.unsqueeze(-1) == labels.unsqueeze(-2)

        # Remove matching pairs if either of them come from the negative class
        unique_matrix = labels < 0
        unique_matrix = unique_matrix.unsqueeze(-1) | unique_matrix.unsqueeze(-2)
        match_matrix *= ~unique_matrix

        # Compress the matrix using the adjacenty matrix (no self connections)
        match_matrix = match_matrix[adjmat].float()

        # Compare the match_matrix to the vertx predictions using the BCE loss
        loss = self.loss(pred.squeeze(-1), match_matrix)

        # If reduction is none and have weight labels, weight the loss
        origin_label = self.label.replace("VertexIndex", "OriginLabel")
        weights = self.get_weights(labels_dict[self.input_name][origin_label], adjmat)
        weighted_loss = loss * weights

        # Calculate the number of non-masked elements
        num_non_masked_elements = match_matrix.sum()

        # Take average over the non-masked elements
        loss = weighted_loss.sum() / num_non_masked_elements

        return loss * self.weight

    def get_weights(self, labels, adjmat):
        weights = torch.clip(sum(labels == i for i in (3, 4, 5)), 0, 1) - (labels == 1).int()
        weights = weights.unsqueeze(-1) & weights.unsqueeze(-2)
        weights = weights[adjmat]
        return 1 + weights

    def run_inference(self, preds: Tensor, pad_mask: Tensor | None = None) -> Tensor:
        preds = get_node_assignment_jit(preds, pad_mask)
        return mask_fill_flattened(preds, pad_mask)

    def get_h5(self, preds: Tensor, pad_mask: Tensor | None = None) -> np.ndarray:
        preds = self.run_inference(preds, pad_mask)
        dtype = np.dtype([("VertexIndex", "i8")])
        return u2s(preds.int().cpu().numpy(), dtype)


# convert flattened array to shape of mask (ntracks, ...) -> (njets, maxtracks, ...)
@torch.jit.script
def mask_fill_flattened(flat_array, mask):
    filled = torch.full((mask.shape[0], mask.shape[1], flat_array.shape[1]), float("-inf"))
    mask = mask.to(torch.bool)
    start_index = end_index = 0

    for i in range(mask.shape[0]):
        if mask[i].shape[0] > 0:
            end_index += (~mask[i]).to(torch.long).sum()
            filled[i, : end_index - start_index] = flat_array[start_index:end_index]
            start_index = end_index

    return filled



import torch
from torch import BoolTensor, Tensor
from torch.nn.functional import pad, softmax

from salt.stypes import Tensors


def maybe_flatten_tensors(
    x: Tensor | Tensors,
) -> Tensor:
    if isinstance(x, dict):
        return flatten_tensor_dict(x)
    return x


def flatten_tensor_dict(
    x: dict[str, Tensor],
    include: list[str] | None = None,
    exclude: list[str] | None = None,
) -> Tensor:
    """Flattens a dictionary of tensors into a single tensor.

    Parameters
    ----------
    x: dict[str, Tensor]
        Dictionary of tensors to flatten.
    include: list[str] | None, optional
        List of keys defining the tensors to be concatenated. If None, all tensors will be
        concatenated unless defined by 'exclude'. Cannot be used with 'exclude'.
    exclude: list[str] | None, optional
        List of keys to exclude from the concatenation. If None, all tensors will be
        concatenated unless defined by 'include'. Cannot be used with 'include'.

    Returns
    -------
    flattened : Tensor
        A single tensor containing the concatenated values of the input dictionary.
    """
    if include and exclude:
        raise ValueError("Cannot use 'include' and 'exclude' together")
    if include:
        return torch.cat([x[emb] for emb in include], dim=1)
    if exclude:
        return torch.cat([x[emb] for emb in x if emb not in exclude], dim=1)
    return torch.cat(list(x.values()), dim=1)


def masked_softmax(x: Tensor, mask: BoolTensor, dim: int = -1) -> Tensor:
    """Applies softmax over a tensor without including padded elements."""
    if mask is not None:
        mask = add_dims(mask, x.dim())
        x = x.masked_fill(mask, -torch.inf)

    x = softmax(x, dim=dim)

    if mask is not None:
        x = x.masked_fill(mask, 0)

    return x


def undo_padding(seq: Tensor, mask: BoolTensor) -> tuple:
    """Remove all padded elements from a tensor and return the sequence lengths."""
    mask = ~mask  # convert the mask such that True is a valid token
    seqlens = mask.sum(dim=-1)
    maxlen = seqlens.max().item()
    culens = pad(torch.cumsum(seqlens, dim=0, dtype=torch.int32), (1, 0))
    return seq[mask], culens, maxlen


def redo_padding(unpadded_seq: Tensor, mask: BoolTensor) -> Tensor:
    """Redo the padding and return a zero-padded tensor."""
    mask = ~mask  # convert the mask such that True is a valid token
    shape = (*mask.shape, unpadded_seq.shape[-1])
    out = torch.zeros(shape, dtype=unpadded_seq.dtype, device=unpadded_seq.device)
    out[mask] = unpadded_seq
    return out


def add_dims(x: Tensor, ndim: int):
    """Adds dimensions to a tensor to match the shape of another tensor."""
    if (dim_diff := ndim - x.dim()) < 0:
        raise ValueError(f"Target ndim ({ndim}) is smaller than input ndim ({x.dim()})")

    if dim_diff > 0:
        x = x.view(x.shape[0], *dim_diff * (1,), *x.shape[1:])

    return x


def attach_context_single(x: Tensor, context: Tensor) -> Tensor:
    if context is None:
        raise RuntimeError("Expected context is missing from forward pass")

    if (dim_diff := x.dim() - context.dim()) < 0:
        raise ValueError(
            f"Provided context has more dimensions ({context.dim()}) than inputs ({x.dim()})"
        )

    if dim_diff > 0:
        context = add_dims(context, x.dim())
        context = context.expand(*x.shape[:-1], -1)

    return torch.cat([context, x], dim=-1)


def attach_context(x: Tensor, context: Tensor) -> Tensor:
    """Concatenate a context tensor to an input tensor with broadcasting.

    The idea behind this is to allow a context tensor less or equal dimensions to be
    concatenated to an input with more dimensions.
    This function checks the dimension difference and reshapes the context to have
    the same dimension as the constituents so broadcast concatenation can apply.
    The shape change always assumes that the first dimension is the batch and the last
    are the features.

    Here is a basic use case: concatenating the high level jet variables to the
    constituents during a forward pass through the network
    - The jet variables will be of shape: [batch, j_features]
    - The constituents will be of shape: [batch, num_nodes, n_features]
    Here the jet variable will be shaped to [batch, 1, j_features] allowing broadcast
    concatenation with the constituents

    Another example is using the edge features [b,n,n,ef] and concatenating the
    high level jet variables, which will be expanded to [b,1,1,jf] or the conditioning
    on the node features which will be expanded to [b,1,n,nf]
    """
    if isinstance(x, dict):
        return {key: attach_context_single(val, context) for key, val in x.items()}

    return attach_context_single(x, context)



import numpy as np
import onnxruntime as ort
import torch
from tqdm import tqdm

from salt.models.task import mask_fill_flattened
from salt.utils.get_structured_input_dict import get_structured_input_dict
from salt.utils.inputs import (
    inputs_sep_with_pad_multi_sequece,
)
from salt.utils.union_find import get_node_assignment_jit

torch.manual_seed(42)


def compare_output(
    pt_model,
    onnx_session,
    global_object,
    seq_names_salt,
    seq_names_onnx,
    variable_map,
    tasks_to_output,
    n_seq=40,
):
    n_batch = 1

    jets, sequences, pad_masks = inputs_sep_with_pad_multi_sequece(
        n_batch,
        [n_seq for seqn in seq_names_salt],
        pt_model.input_dims[global_object],
        [pt_model.input_dims[seqn] for seqn in seq_names_salt],
        p_valid=1,
    )

    inputs_pytorch = {seqn: seq for seq, seqn in zip(sequences, seq_names_salt, strict=False)}
    inputs_pytorch[global_object] = jets

    masks_pytorch = {seqn: mask for mask, seqn in zip(pad_masks, seq_names_salt, strict=False)}

    structured_input_dict = get_structured_input_dict(inputs_pytorch, variable_map, global_object)

    outputs_pytorch = pt_model(inputs_pytorch, masks_pytorch)[0]

    if global_object in outputs_pytorch:
        global_pred_pytorch = []
        global_tasks = [t for t in pt_model.model.tasks if t.input_name == global_object]

        for i, out in enumerate(list(outputs_pytorch[global_object].values())):
            if global_tasks[i].name not in tasks_to_output:
                continue
            onnx_out = global_tasks[i].get_onnx(out, labels=structured_input_dict)
            global_pred_pytorch += [p.detach().numpy() for p in onnx_out]

    else:
        global_pred_pytorch = []

    inputs_onnx = {f"{global_object.removesuffix('s')}_features": jets.numpy()}
    for seq, seqn in zip(sequences, seq_names_onnx, strict=False):
        inputs_onnx[seqn] = seq.squeeze(0).numpy()

    outputs_onnx = onnx_session.run(None, inputs_onnx)

    # test jet classification
    global_pred_onnx = outputs_onnx[: len(global_pred_pytorch)]
    assert not np.isnan(np.array(global_pred_pytorch)).any()
    assert not np.isnan(np.array(global_pred_onnx)).any()
    assert not (np.array(global_pred_pytorch) == 0).any()
    assert not (np.array(global_pred_onnx) == 0).any()
    np.testing.assert_allclose(
        global_pred_pytorch,
        global_pred_onnx,
        rtol=1e-04,
        atol=1e-04,
        err_msg="Torch vs ONNX check failed for global task",
    )

    # Now do the track checks...
    if n_seq == 0:
        return

    # test track origin
    if "track_origin" in tasks_to_output:
        pred_pytorch_origin = (
            torch.argmax(outputs_pytorch["tracks"]["track_origin"], dim=-1).detach().numpy()
        )
        onnx_index = tasks_to_output.index("track_origin") - len(tasks_to_output)
        pred_onnx_origin = outputs_onnx[onnx_index]
        assert (
            len(pred_onnx_origin.shape) == 1
        ), "ONNX output for track origin should be a single tensor"
        np.testing.assert_allclose(
            pred_pytorch_origin.squeeze(),
            pred_onnx_origin,
            rtol=1e-06,
            atol=1e-06,
            err_msg="Torch vs ONNX check failed for track origin",
        )

    # test vertexing
    if "track_vertexing" in tasks_to_output:
        pred_pytorch_scores = outputs_pytorch["tracks"]["track_vertexing"].detach()
        pred_pytorch_indices = get_node_assignment_jit(pred_pytorch_scores, pad_masks[0])
        pred_pytorch_vtx = mask_fill_flattened(pred_pytorch_indices, pad_masks[0])

        onnx_index = tasks_to_output.index("track_vertexing") - len(tasks_to_output)
        pred_onnx_vtx = outputs_onnx[onnx_index]
        np.testing.assert_allclose(
            pred_pytorch_vtx.squeeze(),
            pred_onnx_vtx,
            rtol=1e-06,
            atol=1e-06,
            err_msg="Torch vs ONNX check failed for vertexing",
        )

    # test track type
    if "track_type" in tasks_to_output:
        pred_pytorch_type = (
            torch.argmax(outputs_pytorch["tracks"]["track_type"], dim=-1).detach().numpy()
        )
        onnx_index = tasks_to_output.index("track_type") - len(tasks_to_output)
        pred_onnx_type = outputs_onnx[onnx_index]

        assert (
            len(pred_onnx_type.shape) == 1
        ), "ONNX output for track origin should be a single tensor"
        np.testing.assert_allclose(
            pred_pytorch_type.squeeze(),
            pred_onnx_type,
            rtol=1e-06,
            atol=1e-06,
            err_msg="Torch vs ONNX check failed for track type",
        )


def compare_outputs(
    pt_model,
    onnx_path,
    global_object,
    seq_names_salt,
    seq_names_onnx,
    variable_map,
    tasks_to_output,
):
    print("\n" + "-" * 100)
    print("Validating ONNX model...")

    sess_options = ort.SessionOptions()
    # suppress warnings due to unoptimized subgraphs - https://github.com/microsoft/onnxruntime/issues/14694
    sess_options.log_severity_level = 3
    session = ort.InferenceSession(
        str(onnx_path), providers=["CPUExecutionProvider"], sess_options=sess_options
    )
    for n_track in tqdm(range(40), leave=False):
        for _ in range(10):
            compare_output(
                pt_model,
                session,
                global_object,
                seq_names_salt,
                seq_names_onnx,
                variable_map,
                tasks_to_output,
                n_track,
            )

    print(
        "Success! Pytorch and ONNX models are consistent, but you should verify this in"
        " Athena.\nFor more info see: https://ftag-salt.docs.cern.ch/export/#athena-validation"
    )
    print("-" * 100)



from dataclasses import dataclass


@dataclass
class MaskformerObjectConfig:
    """Sub-config for the Maskformer config, containing information about the object we
    wish to reconstruct.

    Attributes
    ----------
        name (str): Name of the container containing the object
        id_label (str): Name of the variable in the global object to use as an ID
        class_label (str): Name of the variable in the global object to use as a class label
        object_classes: A dictionary of the form:
            class_name_1 (str):
                raw (int): Value that corresponds to value in 'class_label'
                mapped (int): Value to map to (such that all mapped are in {range(num_classes)})
            class_name_2:
                ...
            null:
                mapped: len(object_classes) - 1
    """

    name: str
    id_label: str
    class_label: str | None = None
    object_classes: dict[str, dict[str, int]] | None = None

    def __post_init__(self):
        if self.object_classes:
            # When reading in 'null' jsonparse seems to cast this to 'None', so here we
            # cast it back
            assert None in self.object_classes, "Null class must be present"
            self.object_classes["null"] = self.object_classes.pop(None)

            assert (
                self.object_classes["null"]["mapped"] == len(self.object_classes) - 1
            ), "Null class must be last"
            assert set(self.class_map.values()) == set(range(len(self.object_classes)))

    @property
    def num_classes(self) -> int:
        if not self.object_classes:
            raise ValueError("No object classes defined")
        return len(self.object_classes)

    @property
    def num_not_null_classes(self) -> int:
        if not self.object_classes:
            raise ValueError("No object classes defined")
        return len(self.object_classes) - 1

    @property
    def null_index(self) -> int:
        return self.num_not_null_classes

    @property
    def class_map(self) -> dict[int, int]:
        """Returns a dictionary mapping raw class labels to mapped class labels."""
        if not self.object_classes:
            raise ValueError("No object classes defined")
        return {v["raw"]: v["mapped"] for v in self.object_classes.values()}

    @property
    def class_names(self) -> list[str]:
        """Returns a list of each class name, ordered by the mapped value."""
        if not self.object_classes:
            raise ValueError("No object classes defined")
        sortedmap = sorted(self.object_classes.items(), key=lambda x: x[1]["mapped"])
        return [n for n, _ in sortedmap]


@dataclass
class MaskformerConfig:
    """Configuration for Maskformer model.

    Attributes
    ----------
        objects (MaskformerObjectConfig): Configuration for the objects to be reconstructed
        constituent (MaskformerObjectConfig): Configuration for the constituents of the objects
            The 'class_label' and 'object_classes' attributes are not used here.


    """

    object: MaskformerObjectConfig
    constituent: MaskformerObjectConfig

    def __post_init__(self):
        if isinstance(self.object, dict):
            self.object = MaskformerObjectConfig(**self.object)
        if isinstance(self.constituent, dict):
            self.constituent = MaskformerObjectConfig(**self.constituent)



import torch
from torch import nn
from torch.nn import functional as F


class LayerNorm(nn.LayerNorm):
    def __init__(self, *args, **kwargs):
        """Faster LayerNorm by seting elementwise_affine=False."""
        super().__init__(*args, **kwargs, elementwise_affine=False)


class RMSNorm(nn.Module):
    def __init__(self, dim: int):
        """RNMSNorm from https://arxiv.org/abs/1910.07467. Slower than LayerNorm."""
        super().__init__()
        self.scale = dim**0.5
        self.weight = nn.Parameter(torch.ones(dim))

    def forward(self, x):
        return F.normalize(x, dim=-1) * self.scale * self.weight



from lightning import Callback, LightningModule, Trainer


class WeightLoggerCallback(Callback):
    def __init__(self, log_every_n_steps=50):
        """Callback to log model weights and biases during training.

        Args:
            log_every_n_steps (int): Frequency of logging gradients. Logs every `n` steps.
        """
        self.log_every_n_steps = log_every_n_steps

    def setup(self, trainer: Trainer, module: LightningModule, stage: str) -> None:
        if trainer.fast_dev_run or stage != "fit":
            return
        kwargs = {"sync_dist": len(trainer.device_ids) > 1}

        def log(metrics, stage):
            for t, loss_value in metrics.items():
                n = f"{stage}_{t}"
                module.log(n, loss_value, **kwargs)

        self.log = log

    def on_train_batch_end(self, trainer, pl_module, outputs, batch, batch_idx):  # noqa: ARG002
        """Called after each training batch ends.
        Logs the mean and std of weights and biases in each layer.
        """
        if trainer.global_step % self.log_every_n_steps == 0:
            for name, param in pl_module.named_parameters():
                if "weight" in name or "bias" in name:
                    # Log weight statistics
                    weight_mean = param.data.mean().item()
                    weight_std = param.data.std().item()
                    self.log(
                        {
                            f"weights/{name}_mean": weight_mean,
                            f"weights/{name}_std": weight_std,
                        },
                        "weights",
                    )



import json
from datetime import datetime
from pathlib import Path

from lightning import Callback, LightningModule, Trainer


class PerformanceWriter(Callback):
    def __init__(
        self,
        dir_path: str | None = None,
        add_metrics: list | None = None,
        stdOut: bool = False,
    ) -> None:
        """Write performance metrics to json file.

        Parameters
        ----------
        dir_path : str
            Path to save json file
        add_metrics: list of str
            Optional additional metrics to add to the logs
        stdOut: bool
            Optional: whether to print the performance to stdOut
        """
        super().__init__()
        self.dir_path = dir_path
        self.metrics = ["train_loss", "val_loss", "val_jet_classification_loss"]
        self.metrics += add_metrics if add_metrics is not None else []
        self.stdOut = stdOut

    def setup(self, trainer: Trainer, module: LightningModule, stage: str) -> None:  # noqa: ARG002
        if trainer.fast_dev_run:
            return

        out_dir = Path(trainer.log_dir) if self.dir_path is None else Path(self.dir_path)
        out_dir.mkdir(parents=True, exist_ok=True)
        self.path = Path(out_dir / "performance_metric.json")
        with open(self.path, "w") as _:
            pass

    def on_validation_epoch_end(self, trainer, module) -> None:
        if trainer.state.stage != "validate" or trainer.fast_dev_run:
            return

        with open(self.path, "a") as perf_file:
            timestamp = datetime.now()
            addMetrics = {
                "epoch": module.current_epoch,
                **{
                    metric: f"{trainer.callback_metrics[metric]:.5f}"
                    for metric in self.metrics
                    if metric in trainer.callback_metrics
                },
                "timestamp": timestamp.isoformat(),
            }
            perf_file.write(json.dumps(addMetrics, default=str))
            perf_file.write("\n")  # need for the njson format

            if self.stdOut:
                phraseMetrics = [
                    f"{metric}={trainer.callback_metrics[metric]:.5f}\n"
                    for metric in self.metrics
                    if metric in trainer.callback_metrics
                ]
                content = (
                    f"epoch {module.current_epoch}:\n"
                    f"{''.join(phraseMetrics)}timestamp={timestamp}\n"
                )
                print(content)



import re
from datetime import datetime
from pathlib import Path

import h5py
import numpy as np
import torch
import yaml
from ftag.git_check import check_for_uncommitted_changes, create_and_push_tag
from jsonargparse import Namespace as JsonNamespace
from jsonargparse.typing import register_type
from lightning.pytorch.cli import LightningCLI

from salt.utils.array_utils import listify


# add support for converting yaml lists to tensors
def serializer(x: torch.Tensor) -> list:
    return x.tolist()


def deserializer(x: list) -> torch.Tensor:
    return torch.tensor(x)


register_type(torch.Tensor, serializer, deserializer)


def get_best_epoch(config_path: Path) -> str:
    """Find the best perfoming epoch.

    Parameters
    ----------
    config_path : Path
        Path to saved training config file.

    Returns
    -------
    Path
        Path to best checkpoint for the training run.
    """
    ckpt_dir = Path(config_path.parent / "ckpts")
    print("No --ckpt_path specified, looking for best checkpoint in", ckpt_dir)
    ckpts = list(Path(ckpt_dir).glob("*.ckpt"))
    exp = r"(?<=loss=)(?:(?:\d+(?:\.\d*)?|\.\d+))"
    losses = [float(re.findall(exp, Path(ckpt).name)[0]) for ckpt in ckpts]
    ckpt = ckpts[np.argmin(losses)]
    print("Using checkpoint", ckpt)
    return str(ckpt)


class SaltCLI(LightningCLI):
    def apply_links(self, parser) -> None:
        parser.link_arguments("name", "trainer.logger.init_args.experiment_name")
        parser.link_arguments("name", "model.name")
        parser.link_arguments("trainer.default_root_dir", "trainer.logger.init_args.save_dir")
        parser.link_arguments("data.global_object", "model.global_object")

    def add_arguments_to_parser(self, parser) -> None:
        parser.add_argument("-n", "--name", default="salt", help="Name for this training run.")
        parser.add_argument(
            "-f", "--force", action="store_true", help="Run with uncomitted changes."
        )
        parser.add_argument(
            "-t", "--tag", action="store_true", help="Push a tag for the current code."
        )
        parser.add_argument(
            "--compile", action="store_true", help="Compile the model to speed up training."
        )
        parser.add_argument(
            "-oc", "--overwrite_config", action="store_true", help="Overwrite config file."
        )
        parser.add_argument(
            "-ls",
            "--log_suffix",
            default=None,
            type=str,
            help="Appended to model name to create the log directory.",
        )
        self.apply_links(parser)

    def fit(self, model, **kwargs):
        if self.config[self.subcommand]["compile"]:
            # unfortunately compiling in place doesn't work
            # https://github.com/pytorch/pytorch/pull/97565
            # https://github.com/pytorch/pytorch/issues/101107
            model.model = torch.compile(model.model)
        self.trainer.fit(model, **kwargs)

    def before_instantiate_classes(self) -> None:
        """A lot of automatic configuration is done here."""
        config = self.config[self.subcommand] if self.subcommand else self.config
        sc_tasks = config.model.model.init_args.tasks.init_args.modules

        labels: dict = {}
        for task in sc_tasks:
            assert "Task" in task["class_path"]
            self.collect_labels_from_task(
                task["init_args"], config.model.model.init_args.get("merge_dict"), labels
            )

        if config.model.model.init_args.get("mask_decoder"):
            if not (maskformer_config := config.data.get("mf_config")):
                raise ValueError("Mask decoder requires 'mf_config' in data config.")
            if maskformer_config.constituent.name not in labels:
                raise ValueError(
                    f"The constituent name {maskformer_config.constituent.name} is not in the"
                    " data labels. Ensure that the constituent name is in the input_map of the"
                    " data config."
                )

            # Needed in case no tasks other than mask prediction/classification
            if "objects" not in labels:
                labels["objects"] = []
            labels["objects"] += [
                maskformer_config.object.id_label,
                maskformer_config.object.class_label,
            ]
            labels[maskformer_config.constituent.name] += [maskformer_config.constituent.id_label]

        config.data.labels = labels

        # add norm
        config.model.norm_config = {
            "norm_dict": config.data.norm_dict,
            "variables": config.data.variables,
            "global_object": config.data.global_object,
            "input_map": config.data.input_map,
        }

        if self.subcommand == "fit" or self.subcommand is None:
            # add variables to inititialization networks
            for init_net in config.model.model.init_args.init_nets:
                init_net["variables"] = config.data.variables
                init_net["global_object"] = config.data.global_object

            # add variables to feature-wise networks
            if config.model.model.init_args.featurewise_nets:
                for featurewise_net in config.model.model.init_args.featurewise_nets:
                    featurewise_net["variables"] = config.data.variables

            for task in sc_tasks:
                # extract object class names from h5 attrs (requires FTAG preprocessing)
                if class_names := self.get_object_class_names(task, config.data):
                    task.init_args.class_names = class_names

                # if class weights are not specified read them from class_dict
                if task.init_args.get("use_class_dict") and (
                    class_weights := self.get_class_weights_from_class_dict(task, config.data)
                ):
                    task.init_args.loss.init_args.weight = torch.Tensor(class_weights)

            # reduce precision to improve performance
            # don't do this during evaluation as you will get increased variation wrt Athena
            torch.set_float32_matmul_precision("medium")

            output_dir_path = self.get_output_dir_path(
                config.trainer.default_root_dir, config.name, config.log_suffix
            )
            config.trainer.default_root_dir = output_dir_path
            if config.trainer.logger:
                config.trainer.logger.init_args.save_dir = output_dir_path
                # Create the directory if it doesn't exist
                # Bug fix for comet logger in offline mode
                Path(config.trainer.logger.init_args.save_dir).mkdir(parents=True, exist_ok=True)

            # run git checks
            if not config.force and not config.trainer.fast_dev_run:
                path = Path(__file__).parent
                check_for_uncommitted_changes(path)
                if config.tag:
                    create_and_push_tag(
                        path,
                        "atlas-flavor-tagging-tools/algorithms/salt",
                        Path(output_dir_path).stem,
                        "automated salt tag",
                    )

            if config.overwrite_config:
                self.save_config_kwargs["overwrite"] = True

        if self.subcommand == "test":
            print("\n" + "-" * 100)

            # no logger, callback refresh rate 1 for testing
            self.save_config_callback = None
            config.trainer.logger = False
            for callback in config.trainer.callbacks:
                if hasattr(callback, "init_args") and hasattr(callback.init_args, "refresh_rate"):
                    callback.init_args.refresh_rate = 1

            # use best epoch for testing
            if not config.ckpt_path:
                assert len(config.config) == 1
                config.ckpt_path = get_best_epoch(Path(config.config[0].rel_path))

            if isinstance(config.trainer.devices, str | int) and int(config.trainer.devices) > 1:
                print("Setting --trainer.devices=1")
                config.trainer.devices = "1"
            elif isinstance(config.trainer.devices, list) and len(config.trainer.devices) > 1:
                raise ValueError("Testing requires --trainer.devices=1")

            config.data.move_files_temp = None

            print("-" * 100 + "\n")

    @staticmethod
    def collect_labels_from_task(
        task: JsonNamespace,
        merge_dict: JsonNamespace | None,
        labels: dict,
    ) -> None:
        def collect_labels(task: JsonNamespace):
            labels = []
            if label := task.get("label"):
                labels.append(label)
            if weight := task.get("sample_weight"):
                labels.append(weight)
            if targets := task.get("targets"):
                labels.extend(listify(targets))

            if denominators := task.get("target_denominators"):
                labels.extend(listify(denominators))

            return labels

        if task["input_name"] in labels:
            labels[task["input_name"]].extend(collect_labels(task))
        else:
            labels[task["input_name"]] = collect_labels(task)

        if not merge_dict or task["input_name"] not in merge_dict:
            return

        for label_name in merge_dict[task["input_name"]]:
            if label_name in labels:
                labels[label_name].extend(collect_labels(task))
            else:
                labels[label_name] = collect_labels(task)

    @staticmethod
    def get_object_class_names(task: JsonNamespace, data: JsonNamespace) -> list | None:
        if not (
            task.init_args.name == f"{data.global_object}_classification"
            and task.init_args.label == "flavour_label"
            and task.init_args.class_names is None
        ):
            return None

        name = (
            data.input_map[task.init_args.input_name]
            if data.input_map
            else task.init_args.input_name
        )
        with h5py.File(data.train_file) as f:
            if task.init_args.label not in f[name].attrs:
                raise ValueError(
                    f"'{task.init_args.label}' not found in the h5 attrs of group '{name}' in file "
                    f"{data.train_file}. Specify class_names manually."
                )

            return list(f[name].attrs[task.init_args.label])

    @staticmethod
    def get_class_weights_from_class_dict(
        task: JsonNamespace,
        data: JsonNamespace,
    ) -> list:
        if (class_dict_path := data.class_dict) is None:
            raise ValueError("use_class_dict=True requires class_dict to be specified")

        if task.init_args.loss.init_args.weight is not None:
            raise ValueError(
                "Class weights already specified, disable use_class_dict or remove weights."
            )

        with open(class_dict_path) as f:
            class_dict = yaml.safe_load(f)

        input_name = task.init_args.input_name
        if data.input_map is not None:
            input_name = data.input_map[input_name]

        if task.init_args.label not in class_dict[input_name]:
            raise ValueError(
                f"Label {task.init_args.label} not found in class_dict. "
                "Use use_class_dict=False and specify class weights manually."
            )

        return class_dict[input_name][task.init_args.label]

    @staticmethod
    def get_output_dir_path(default_output_dir_path: str, name: str, log_suffix: str | None) -> str:
        log_dir = Path(default_output_dir_path)
        try:
            datetime.strptime(log_dir.name.split("_")[-1], "%Y%m%d-T%H%M%S")
            log_dir = log_dir.parent
        except ValueError:
            pass

        if log_suffix is not None:
            dirname = f"{name}_{log_suffix}"
        else:
            timestamp = datetime.now().strftime("%Y%m%d-T%H%M%S")
            dirname = f"{name}_{timestamp}"

        return str(Path(log_dir / dirname).resolve())



import numpy as np


class GaussianNoise:
    def __init__(self, noise_params=None):
        """Initializes the Gaussian noise generator with customizable parameters.

        Parameters
        ----------
        noise_params : dict, optional
            A dictionary where keys are input names (e.g., "tracks", "jets") and values are
            dictionaries with "mean", "std", and optionally a "mask" array for selective noise.
        """
        self.noise_params = noise_params if noise_params is not None else {}

    def add_noise(self, data, mean, std):
        """Add Gaussian noise to specified subfields of a dataset.

        Parameters
        ----------
        data : Numpy array
            Array containing data to which noise will be added
        mean : float
            Mean of noise to be added
        std: : float
            Standard deviation of noise to be added

        Returns
        -------
        data : Numpy array
            Data with noise added
        """
        # Generate noise
        rng = np.random.default_rng()
        noise = rng.normal(mean, std, data.shape)

        # Scale noise based on data and apply it
        data_noise = data * noise
        data += data_noise

        return data

    def __call__(self, data, input_type):
        """Applies Gaussian noise to all variables of the given type as specified in `noise_params`.

        Parameters
        ----------
        data : Numpy array
            Array containing data for input variables
        input_type : str
            Name of the input variable type e.g. jets or tracks

        Returns
        -------
        data : Numpy array
            The modified data array with noise applied
        """
        for noise_dict in self.noise_params:
            # Skip if wrong input type
            if noise_dict["input_type"] != input_type:
                continue
            var = noise_dict["variable"]
            mean = noise_dict["mean"]
            std = noise_dict["std"]
            noise_data = data[var]
            data[var] = self.add_noise(noise_data, mean, std)
        return data



import importlib.util

import pytest
import torch
from torch import nn
from torch.utils.benchmark import Timer

from salt.models.attention import MultiheadAttention
from salt.models.layernorm import RMSNorm
from salt.models.transformer_v2 import (
    Attention,
    DecoderLayer,
    TransformerV2,
    change_attn_backends,
    merge_masks,
    redo_padding,
    undo_padding,
)

N_BATCH = 10
Q_SEQ = 20
KV_SEQ = 10
DIM = 16


def create_bool_tensor(shape, value):
    return torch.full(shape, value, dtype=torch.bool)


def test_merge_masks_none_inputs():
    q_shape = (N_BATCH, Q_SEQ, DIM)
    mask = merge_masks(None, None, q_shape)
    assert mask is None


def test_merge_masks_only_attn_mask():
    q_shape = (N_BATCH, Q_SEQ, DIM)
    attn_shape = (N_BATCH, Q_SEQ, KV_SEQ)
    attn_mask = create_bool_tensor(attn_shape, False)
    mask = merge_masks(None, attn_mask, q_shape)
    assert mask.shape == (N_BATCH, 1, Q_SEQ, KV_SEQ)


def test_merge_masks_only_kv_mask():
    q_shape = (N_BATCH, Q_SEQ, DIM)
    k_shape = (N_BATCH, KV_SEQ, DIM)
    kv_mask = create_bool_tensor(k_shape[:-1], False)
    mask = merge_masks(kv_mask, None, q_shape)
    assert mask.shape == (N_BATCH, 1, Q_SEQ, KV_SEQ)


def test_merge_masks_attn_and_kv_masks():
    q_shape = (N_BATCH, Q_SEQ, DIM)
    k_shape = (N_BATCH, KV_SEQ, DIM)
    attn_shape = (N_BATCH, Q_SEQ, KV_SEQ)
    kv_mask = create_bool_tensor(k_shape[:-1], False)
    attn_mask = create_bool_tensor(attn_shape, True)
    mask = merge_masks(kv_mask, attn_mask, q_shape)
    assert mask.shape == (N_BATCH, 1, Q_SEQ, KV_SEQ)
    assert torch.all(mask)


def test_padding_mask():
    torch_attn = nn.MultiheadAttention(8, 1, batch_first=True)

    # this is a correct full attention mask for padded inputs
    # we only need to pad the keys so that no query receives information from them
    # if we were to also pad the queries, the padded tokens would become nan as
    # they would not recieve any incomming signal
    attn_mask = torch.tensor([[[False, False, True], [False, False, True], [False, False, True]]])

    # confirm that the value of the padded tokens has no effect on the output
    x = torch.ones(1, 3, 8)
    x[:, 2] = 0
    out1 = torch_attn(x, x, x, attn_mask=attn_mask)[0]
    x[:, 2] = 10
    out2 = torch_attn(x, x, x, attn_mask=attn_mask)[0]
    torch.testing.assert_close(out1, out2)

    # this kind of mask is overkill and leads to nans
    # attn_mask = torch.tensor([[
    #   [False, False,  True],
    #   [False, False,  True],
    #   [ True,  True,  True]
    # ]])


def get_models(dim, num_heads) -> tuple:
    salt_attn = Attention(dim, num_heads=num_heads)
    torch_attn = nn.MultiheadAttention(dim, num_heads, batch_first=True)
    salt_attn.in_proj_weight = torch_attn.in_proj_weight
    salt_attn.in_proj_bias = torch_attn.in_proj_bias
    salt_attn.out_proj.weight = torch_attn.out_proj.weight
    salt_attn.out_proj.bias = torch_attn.out_proj.bias
    return salt_attn, torch_attn


def get_cross_attn_inputs(batch_size, q_len, kv_len, dim, frac_pad=0.0) -> tuple:
    torch.manual_seed(0)
    q = torch.randn(batch_size, q_len, dim)
    kv = torch.randn(batch_size, kv_len, dim)
    kv_mask = torch.rand(batch_size, kv_len) > frac_pad
    kv_mask[:, 0] = False  # Make sure something can send
    return q, kv, kv_mask


def get_self_attn_inputs(batch_size, seq_len, dim, frac_pad=0.0) -> tuple:
    torch.manual_seed(0)
    x = torch.randn(batch_size, seq_len, dim)
    mask = torch.rand(batch_size, seq_len) > frac_pad
    mask[:, 0] = False  # Make sure something can send
    return x, mask


@pytest.mark.parametrize("batch_size", [1, 10])
@pytest.mark.parametrize("q_len", [1, 10])
@pytest.mark.parametrize("kv_len", [1, 10])
@pytest.mark.parametrize("dim", [32])
@pytest.mark.parametrize("frac_pad", [0.0, 0.5, 0.9])
def test_cross_attention(
    batch_size,
    q_len,
    kv_len,
    dim,
    frac_pad,
) -> None:
    salt_attn, torch_attn = get_models(dim, 2)
    q, kv, kv_mask = get_cross_attn_inputs(batch_size, q_len, kv_len, dim, frac_pad)
    custom_output = salt_attn(q, kv, kv_mask=kv_mask)
    torch_output, _ = torch_attn(q, kv, kv, key_padding_mask=kv_mask)
    torch.testing.assert_close(custom_output, torch_output)
    assert not torch.isnan(custom_output).any()


@pytest.mark.parametrize("batch_size", [1, 10])
@pytest.mark.parametrize("seq_len", [1, 2, 10])
@pytest.mark.parametrize("dim", [32])
@pytest.mark.parametrize("num_heads", [1, 2])
@pytest.mark.parametrize("frac_pad", [0.0, 0.5, 0.9])
def test_self_attention(
    batch_size,
    seq_len,
    dim,
    num_heads,
    frac_pad,
) -> None:
    salt_attn, torch_attn = get_models(dim, num_heads)
    x, mask = get_self_attn_inputs(batch_size, seq_len, dim, frac_pad)
    custom_output = salt_attn(x, mask=mask)
    torch_output, _ = torch_attn(x, x, x, key_padding_mask=mask)
    torch.testing.assert_close(custom_output, torch_output)
    assert not torch.isnan(custom_output).any()


@pytest.mark.parametrize("batch_size", [1, 10])
@pytest.mark.parametrize("seq_len", [1, 2, 10])
@pytest.mark.parametrize("dim", [32])
@pytest.mark.parametrize("num_heads", [1, 2])
@pytest.mark.parametrize("frac_pad", [0.0, 0.5])
@pytest.mark.parametrize("attn_type", ["torch-flash", "torch-meff", "flash-varlen"])
def test_attention_backends(
    batch_size,
    seq_len,
    dim,
    num_heads,
    frac_pad,
    attn_type,
) -> None:
    if not torch.cuda.is_available():
        pytest.skip("CUDA not available")
    if importlib.util.find_spec("flash_attn") is None:
        pytest.skip("flash_attn not available")

    # FlashVarlenAttention requires half precision
    with torch.autocast("cuda", enabled=True):
        # Get the inputs and move to device
        x, mask = get_self_attn_inputs(batch_size, seq_len, dim, frac_pad)
        x = x.cuda()
        mask = mask.cuda()

        # Change the masking to None for the torch backends as they dont support it
        if "torch" in attn_type:
            mask = None

        # Perform the standard attention (math)
        attn = Attention(dim, num_heads=num_heads).to("cuda")
        output = attn(x, mask=mask)

        # ensure zero padded
        if mask is not None:
            output *= ~mask.unsqueeze(-1)

        # Switch to the attention backend
        attn.set_backend(attn_type)
        if attn_type == "flash-varlen":
            x_p, culens, maxlen = undo_padding(x, mask)
            output_2 = attn(x_p, mask=mask, culens=culens, maxlen=maxlen)
            output_2 = redo_padding(output_2, mask)
        else:
            output_2 = attn(x, mask=mask)

        # Test all close with less strict due to half precision
        torch.testing.assert_close(output, output_2, atol=1e-3, rtol=1e-3)
        assert not torch.isnan(output_2).any()


def sync_v1v2_attn(v1_attn, v2_attn):
    wq, wk, wv = v2_attn.in_proj_weight.chunk(3)
    bq, bk, bv = v2_attn.in_proj_bias.chunk(3)
    v1_attn.linear_q.weight.data = wq
    v1_attn.linear_k.weight.data = wk
    v1_attn.linear_v.weight.data = wv
    v1_attn.linear_q.bias.data = bq
    v1_attn.linear_k.bias.data = bk
    v1_attn.linear_v.bias.data = bv


@pytest.mark.parametrize("dim", [32])
@pytest.mark.parametrize("num_heads", [1, 2])
@pytest.mark.parametrize("frac_pad", [0.0, 0.5, 0.9])
def test_v1_v2_attention_output(dim, num_heads, frac_pad):
    v1_attn = MultiheadAttention(
        dim, num_heads, {"class_path": "salt.models.ScaledDotProductAttention"}
    )
    v2_attn = Attention(dim, num_heads=num_heads)
    sync_v1v2_attn(v1_attn, v2_attn)
    v1_attn.linear_out = v2_attn.out_proj
    q, kv, kv_mask = get_cross_attn_inputs(10, 20, 20, dim, frac_pad=frac_pad)
    v1_out = v1_attn(q, kv, kv_mask=kv_mask)
    v2_out = v2_attn(q, kv, kv_mask=kv_mask)
    torch.testing.assert_close(v1_out, v2_out)


@pytest.mark.parametrize("num_registers", [1, 4])
@pytest.mark.parametrize("num_layers", [1, 3])
@pytest.mark.parametrize("ls_init", [None, 0.1])
@pytest.mark.parametrize("drop_path", [0, 0.1])
def test_transformerv2_tensor_input(num_registers, num_layers, ls_init, drop_path):
    x, mask = get_self_attn_inputs(5, 10, 32, 0.5)
    trans = TransformerV2(
        num_layers=num_layers,
        embed_dim=32,
        attn_type="torch-math",
        dense_kwargs={"activation": "SiLU"},
        attn_kwargs={"num_heads": 2},
        num_registers=num_registers,
        ls_init=ls_init,
        drop_path=drop_path,
    )
    x, mask = trans(x, pad_mask=mask)
    assert x.shape == (5, 10 + num_registers, 32)
    assert not x.isnan().any()


@pytest.mark.parametrize("ls_init", [None, 0.1])
@pytest.mark.parametrize("drop_path", [0, 0.1])
def test_decoder_layer(ls_init, drop_path):
    q, kv, kv_mask = get_cross_attn_inputs(5, 10, 5, 32, 0.5)
    decoder = DecoderLayer(
        embed_dim=32,
        dense_kwargs={"activation": "SiLU"},
        attn_kwargs={"num_heads": 2},
        ls_init=ls_init,
        drop_path=drop_path,
    )
    x = decoder(q, kv=kv, kv_mask=kv_mask)
    assert x.shape == q.shape
    assert not x.isnan().any()


@pytest.mark.parametrize("num_registers", [1, 4])
def test_transformerv2_dict_input(num_registers):
    x1, m1 = get_self_attn_inputs(5, 10, 32, 0.5)
    x2, m2 = get_self_attn_inputs(5, 3, 32, 0.5)
    x3, m3 = get_self_attn_inputs(5, 2, 32, 0.5)
    x = {"m1": x1, "m2": x2, "m3": x3}  # Multimodal inputs
    mask = {"m1": m1, "m2": m2, "m3": m3}
    trans = TransformerV2(
        num_layers=3,
        embed_dim=32,
        attn_type="torch-math",
        dense_kwargs={"activation": "SiLU"},
        attn_kwargs={"num_heads": 2},
        num_registers=num_registers,
    )
    x, mask = trans(x, pad_mask=mask)
    assert x.shape == (5, 10 + 3 + 2 + num_registers, 32)
    assert all(k in mask for k in ["m1", "m2", "m3", "REGISTERS"])


def test_times_torch_vs_salt() -> None:
    # skip if cuda is not available
    if not torch.cuda.is_available():
        pytest.skip("CUDA not available")

    # Define the input parameters for the timings
    batch_size, seq_len, dim, num_heads = 1000, 64, 128, 8
    salt_attn, torch_attn = get_models(dim, num_heads)
    x, mask = get_self_attn_inputs(batch_size, seq_len, dim, frac_pad=0.5)

    # move tensors and models to cuda
    x = x.cuda()
    mask = mask.cuda()
    salt_attn.cuda()
    torch_attn.cuda()

    # avoid torch fast path
    salt_attn.training = True
    torch_attn.training = True

    # Using timers also performs warm up
    salt_timer = Timer(
        stmt="salt_attn(x, kv_mask=mask)",
        globals={"salt_attn": salt_attn, "x": x, "mask": mask},
        label="salt",
        num_threads=1,
    )

    torch_timer = Timer(
        stmt="torch_attn(x, x, x, key_padding_mask=mask)",
        globals={"torch_attn": torch_attn, "x": x, "mask": mask},
        label="torch",
        num_threads=1,
    )

    salt_time = salt_timer.timeit(300).mean
    torch_time = torch_timer.timeit(300).mean
    assert salt_time < torch_time, f"mean: {salt_time} vs {torch_time}"


def test_times_varlen_vs_default() -> None:
    if not torch.cuda.is_available():
        pytest.skip("CUDA not available")
    if importlib.util.find_spec("flash_attn") is None:
        pytest.skip("flash_attn not available")

    # FlashVarlenAttention requires half precision
    with torch.autocast("cuda", enabled=True):
        # Define the input parameters for the timings
        num_layers = 4
        num_heads = 4
        batch_size = 256
        seq_len = 64
        dim = 128
        x, mask = get_self_attn_inputs(batch_size, seq_len, dim, frac_pad=0.5)

        # Create the transformers
        standard_attn = TransformerV2(
            num_layers=num_layers,
            embed_dim=dim,
            attn_type="torch-math",
            dense_kwargs={"activation": "SiLU"},
            attn_kwargs={"num_heads": num_heads},
        )

        varlen_attn = TransformerV2(
            num_layers=num_layers,
            embed_dim=dim,
            attn_type="flash-varlen",
            dense_kwargs={"activation": "SiLU"},
            attn_kwargs={"num_heads": num_heads},
        )

        # move tensors and models to cuda
        x = x.cuda()
        mask = mask.cuda()
        standard_attn.cuda()
        varlen_attn.cuda()

        # Time the models
        s_timer = Timer(
            stmt="standard_attn(x, pad_mask=mask)",
            globals={"standard_attn": standard_attn, "x": x, "mask": mask},
            label="salt",
            num_threads=1,
        )
        v_timer = Timer(
            stmt="varlen_attn(x, pad_mask=mask)",
            globals={"varlen_attn": varlen_attn, "x": x, "mask": mask},
            label="salt",
            num_threads=1,
        )
        st = s_timer.timeit(20).mean
        vt = v_timer.timeit(20).mean
        assert vt < st, f"mean: {vt} vs {st}"


def test_RMSNorm():
    rmsnorm = RMSNorm(10)
    x = torch.randn(5, 10)
    rmsnorm(x)


def test_DecoderLayer():
    layer = DecoderLayer(embed_dim=32, attn_kwargs={"num_heads": 2})
    x = torch.randn(5, 10, 32)
    kv = torch.randn(5, 10, 32)
    layer(x, kv=kv)


def test_change_attn_backends():
    model = TransformerV2(
        num_layers=3,
        embed_dim=32,
        attn_type="torch-math",
        dense_kwargs={"activation": "SiLU"},
        attn_kwargs={"num_heads": 2},
    )

    # change the backend
    change_attn_backends(model, "torch-meff")
    assert model.attn_type == "torch-meff"
    for layer in model.layers:
        assert layer.attn.fn.attn_type == "torch-meff"

    # no cuda, so it should not be able to set flash-varlen, and isntead fall back to torch-math
    if not torch.cuda.is_available():
        with pytest.warns(UserWarning):
            change_attn_backends(model, "flash-varlen")
        assert model.attn_type == "torch-math"
        for layer in model.layers:
            assert layer.attn.fn.attn_type == "torch-math"

    # check this works for a module that wraps a transformer
    wrapper = nn.Sequential(model)
    change_attn_backends(wrapper, "torch-flash")
    assert model.attn_type == "torch-flash"
    for layer in model.layers:
        assert layer.attn.fn.attn_type == "torch-flash"

    # check this works for a base attention layer
    attn = Attention(32, num_heads=2, attn_type="torch-math")
    change_attn_backends(attn, "torch-flash")
    assert attn.attn_type == "torch-flash"



"""Mostly helper functions for generating random inputs."""

from pathlib import Path

import h5py
import numpy as np
import torch
import yaml
from numpy.lib.recfunctions import unstructured_to_structured as u2s
from numpy.random import default_rng

from salt.utils.array_utils import join_structured_arrays
from salt.utils.tensor_utils import attach_context

DEFAULT_NTRACK = 40

JET_VARS = [
    "pt",
    "eta",
    "mass",
    "pt_btagJes",
    "eta_btagJes",
    "softMuon_pt",
    "softMuon_dR",
    "softMuon_eta",
    "softMuon_phi",
    "softMuon_qOverPratio",
    "softMuon_momentumBalanceSignificance",
    "softMuon_scatteringNeighbourSignificance",
    "softMuon_pTrel",
    "softMuon_ip3dD0",
    "softMuon_ip3dZ0",
    "softMuon_ip3dD0Significance",
    "softMuon_ip3dZ0Significance",
    "softMuon_ip3dD0Uncertainty",
    "softMuon_ip3dZ0Uncertainty",
    "R10TruthLabel_R22v1",
    "R10TruthLabel_R22v1_TruthJetMass",
    "R10TruthLabel_R22v1_TruthJetPt",
    "GN2Xv00_phbb",
    "GN2Xv00_phcc",
    "GN2Xv00_ptop",
    "GN2Xv00_pqcd",
    "GN2XWithMassv00_phbb",
    "GN2XWithMassv00_phcc",
    "GN2XWithMassv00_ptop",
    "GN2XWithMassv00_pqcd",
    "Xbb2020v3_Higgs",
    "Xbb2020v3_Top",
    "Xbb2020v3_QCD",
]

TRACK_VARS = [
    "d0",
    "z0SinTheta",
    "dphi",
    "deta",
    "qOverP",
    "IP3D_signed_d0_significance",
    "IP3D_signed_z0_significance",
    "phiUncertainty",
    "thetaUncertainty",
    "qOverPUncertainty",
    "numberOfPixelHits",
    "numberOfSCTHits",
    "numberOfInnermostPixelLayerHits",
    "numberOfNextToInnermostPixelLayerHits",
    "numberOfInnermostPixelLayerSharedHits",
    "numberOfInnermostPixelLayerSplitHits",
    "numberOfPixelSharedHits",
    "numberOfPixelSplitHits",
    "numberOfSCTSharedHits",
    "numberOfPixelHoles",
    "numberOfSCTHoles",
    "pt",
    "eta",
    "phi",
    "subjetIndex",
    "leptonID",
]

ELECTRON_VARS = [
    "pt",
    "ptfrac",
    "ptrel",
    "dr",
    "abs_eta",
    "eta",
    "phi",
    "ftag_et",
    "qOverP",
    "d0RelativeToBeamspotSignificance",
    "ftag_z0AlongBeamspotSignificance",
    "ftag_ptVarCone30OverPt",
    "numberOfPixelHits",
    "numberOfSCTHitsInclDead",
    "ftag_deltaPOverP",
    "eProbabilityHT",
    "deltaEta1",
    "deltaPhiRescaled2",
    "ftag_energyOverP",
    "Rhad",
    "Rhad1",
    "Eratio",
    "weta2",
    "Rphi",
    "Reta",
    "wtots1",
    "f1",
    "f3",
]

HADRON_VARS = ["pt", "Lxy", "deta", "dphi", "mass"]
rng = np.random.default_rng(42)
torch.manual_seed(42)


def inputs_sep_no_pad(n_batch: int, n_track: int, n_jet_feat: int, n_track_feat: int):
    jets = torch.rand(n_batch, n_jet_feat)
    tracks = torch.rand(n_batch, n_track, n_track_feat)
    return jets, tracks


def inputs_sep_no_pad_multi_sequece(
    n_batch: int, n_seq_list: list[int], n_jet_feat: int, n_seq_feat_list: list[int]
):
    jets = torch.rand(n_batch, n_jet_feat)
    sequences = []
    for n_seq, n_seq_feat in zip(n_seq_list, n_seq_feat_list, strict=False):
        sequences.append(torch.rand(n_batch, n_seq, n_seq_feat))
    return jets, sequences


def inputs_sep_with_pad(
    n_batch: int, n_track: int, n_jet_feat: int, n_track_feat: int, p_valid=0.5
):
    jets, tracks = inputs_sep_no_pad(n_batch, n_track, n_jet_feat, n_track_feat)
    mask = get_random_mask(n_batch, n_track, p_valid)
    return jets, tracks, mask


def inputs_sep_with_pad_multi_sequece(
    n_batch: int, n_seq_list: list[int], n_jet_feat: int, n_seq_feat_list: list[int], p_valid=0.5
):
    jets, sequences = inputs_sep_no_pad_multi_sequece(
        n_batch, n_seq_list, n_jet_feat, n_seq_feat_list
    )
    masks = [get_random_mask(n_batch, n_seq, p_valid) for n_seq in n_seq_list]
    return jets, sequences, masks


def get_random_mask(n_batch: int, n_track: int, p_valid: float = 0.5):
    a = rng.choice(a=[True, False], size=(n_batch, n_track), p=[1 - p_valid, p_valid])
    a = np.sort(a, axis=-1)[:, ::-1]
    if n_track > 0 and p_valid > 0:  # ensure at least one valid track
        a[:, 0] = False
    return torch.tensor(a.copy())


def inputs_concat(n_batch: int, n_track: int, n_jet_feat: int, n_track_feat: int):
    jets, tracks = inputs_sep_no_pad(n_batch, DEFAULT_NTRACK, n_jet_feat, n_track_feat)
    inputs = attach_context(tracks, jets)
    mask = get_random_mask(n_batch, n_track)
    return inputs, mask


def write_dummy_norm_dict(nd_path: Path, cd_path: Path):
    sd: dict = {}
    sd["jets"] = {n: {"std": 1.0, "mean": 1.0} for n in JET_VARS}
    sd["tracks"] = {n: {"std": 1.0, "mean": 1.0} for n in TRACK_VARS}
    sd["tracks_dr"] = {n: {"std": 1.0, "mean": 1.0} for n in TRACK_VARS}
    sd["tracks_ghost"] = {n: {"std": 1.0, "mean": 1.0} for n in TRACK_VARS}
    sd["electrons"] = {n: {"std": 1.0, "mean": 1.0} for n in ELECTRON_VARS}
    sd["flow"] = {n: {"std": 1.0, "mean": 1.0} for n in TRACK_VARS}
    with open(nd_path, "w") as file:
        yaml.dump(sd, file, sort_keys=False)

    cd: dict = {}
    cd["jets"] = {"HadronConeExclTruthLabelID": [1.0, 2.0, 2.0, 2.0]}
    cd["jets"]["flavour_label"] = cd["jets"]["HadronConeExclTruthLabelID"]
    cd["tracks"] = {"ftagTruthOriginLabel": [4.2, 73.7, 1.0, 17.5, 12.3, 12.5, 141.7, 22.3]}
    cd["tracks_ghost"] = {"ftagTruthOriginLabel": [4.2, 73.7, 1.0, 17.5, 12.3, 12.5, 141.7, 22.3]}
    with open(cd_path, "w") as file:
        yaml.dump(cd, file, sort_keys=False)


def get_dummy_inputs(n_jets=1000, n_jet_features=2, n_track_features=21, n_tracks_per_jet=40):
    shapes_jets = {
        "inputs": [n_jets, n_jet_features],
        "labels": [n_jets],
        "add_labels": [n_jets],
    }

    shapes_tracks = {
        "inputs": [n_jets, n_tracks_per_jet, n_jet_features + n_track_features],
        "labels/ftagTruthOriginLabel": [n_jets, n_tracks_per_jet],
        "labels/ftagTruthParentBarcode": [n_jets, n_tracks_per_jet],
        "labels/ftagTruthVertexIndex": [n_jets, n_tracks_per_jet],
        "labels/truthOriginLabel": [n_jets, n_tracks_per_jet],
        "labels/truthVertexIndex": [n_jets, n_tracks_per_jet],
        "valid": [n_jets, n_tracks_per_jet],
    }

    rng = default_rng(seed=42)
    jets = {}
    for key, shape in shapes_jets.items():
        jets[key] = rng.random(shape)

    tracks = {}
    for key, shape in shapes_tracks.items():
        arr = rng.random(shape)
        if key == "valid":
            arr = arr.astype(bool)
            arr[:, 10:] = False
        if key == "labels":
            arr = rng.integers(0, 8, shape)
        tracks[key] = arr

    return jets, tracks


def write_dummy_file(fname, sd_fname, make_xbb=False, inc_taus=False, inc_params=False):
    """TODO: merge with atlas-ftag-tools test file generation."""
    with open(sd_fname) as f:
        sd = yaml.safe_load(f)

    jet_vars = [
        "pt",
        "eta",
        "mass",
        "pt_btagJes",
        "eta_btagJes",
        "HadronConeExclTruthLabelPt",
        "HadronConeExclTruthLabelLxy",
        "n_tracks",
        "n_truth_promptLepton",
        "softMuon_pt",
        "softMuon_dR",
        "softMuon_eta",
        "softMuon_phi",
        "softMuon_qOverPratio",
        "softMuon_momentumBalanceSignificance",
        "softMuon_scatteringNeighbourSignificance",
        "softMuon_pTrel",
        "softMuon_ip3dD0",
        "softMuon_ip3dZ0",
        "softMuon_ip3dD0Significance",
        "softMuon_ip3dZ0Significance",
        "softMuon_ip3dD0Uncertainty",
        "softMuon_ip3dZ0Uncertainty",
        "R10TruthLabel_R22v1",
        "R10TruthLabel_R22v1_TruthJetMass",
        "R10TruthLabel_R22v1_TruthJetPt",
        "GN2Xv00_phbb",
        "GN2Xv00_phcc",
        "GN2Xv00_ptop",
        "GN2Xv00_pqcd",
        "GN2XWithMassv00_phbb",
        "GN2XWithMassv00_phcc",
        "GN2XWithMassv00_ptop",
        "GN2XWithMassv00_pqcd",
        "Xbb2020v3_Higgs",
        "Xbb2020v3_Top",
        "Xbb2020v3_QCD",
    ]

    params = ["mass"]

    track_vars = list(sd["tracks"])
    electron_vars = list(sd["electrons"])

    # settings
    n_jets = 1000
    jet_features = len(jet_vars)
    n_tracks_per_jet = 40
    track_features = len(track_vars)
    n_electrons_per_jet = 10
    electron_features = len(electron_vars)
    n_hadrons_per_jet = 5
    hadron_features = len(HADRON_VARS)
    # setup jets
    shapes_jets = {
        "inputs": [n_jets, jet_features + 2],
    }

    # setup tracks
    shapes_tracks = {
        "inputs": [n_jets, n_tracks_per_jet, track_features + 3],
        "valid": [n_jets, n_tracks_per_jet],
    }

    shapes_electrons = {
        "inputs": [n_jets, n_electrons_per_jet, electron_features + 2],
        "valid": [n_jets, n_electrons_per_jet],
    }

    shapes_hadrons = {
        "inputs": [n_jets, n_hadrons_per_jet, hadron_features + 2],
        "valid": [n_jets, n_hadrons_per_jet],
    }

    # setup parameters
    shapes_params = {
        "inputs": [n_jets, len(params)],
    }

    # setup jets
    jets_dtype = np.dtype(
        [(n, "f4") for n in jet_vars]
        + [("flavour_label", "i4"), ("HadronConeExclTruthLabelID", "i4")]
    )
    jets = rng.random(shapes_jets["inputs"])
    jets = u2s(jets, jets_dtype)
    if make_xbb:
        jets["flavour_label"] = rng.choice([0, 1, 2, 3], size=n_jets)
    else:
        jets["flavour_label"] = rng.choice([0, 1, 2], size=n_jets)

    jets["HadronConeExclTruthLabelID"] = rng.choice([0, 4, 5], size=n_jets)
    jets["HadronConeExclTruthLabelLxy"][jets["flavour_label"] == 0] = np.nan

    # setup hadrons
    hadrons_dtype = np.dtype(
        [(n, "f4") for n in HADRON_VARS] + [("barcode", "i4"), ("flavour", "i4")]
    )
    hadrons = rng.random(shapes_hadrons["inputs"])
    valid = rng.choice([True, False], size=shapes_hadrons["valid"])
    valid = np.sort(valid, axis=-1)[:, ::-1].view(dtype=np.dtype([("valid", bool)]))
    hadrons[~valid["valid"]] = np.nan
    hadrons = u2s(hadrons, hadrons_dtype)
    hadrons = join_structured_arrays([hadrons, valid])
    hadrons["barcode"] = rng.integers(0, 10000, size=(n_jets, n_hadrons_per_jet))
    hadrons["barcode"][~hadrons["valid"]] = -1
    hadrons["flavour"] = rng.choice([-1, 4, 5], size=(n_jets, n_hadrons_per_jet))
    hadrons["flavour"] = np.sort(hadrons["flavour"], axis=-1)[:, ::-1]
    hadrons["flavour"][~hadrons["valid"]] = -1

    # setup tracks
    tracks_dtype = np.dtype(
        [(n, "f4") for n in track_vars]
        + [
            ("ftagTruthOriginLabel", "i4"),
            ("ftagTruthVertexIndex", "i4"),
            ("ftagTruthParentBarcode", "i4"),
        ]
    )
    tracks = rng.random(shapes_tracks["inputs"])
    valid = rng.choice([True, False], size=shapes_tracks["valid"])
    valid = np.sort(valid, axis=-1)[:, ::-1].view(dtype=np.dtype([("valid", bool)]))
    tracks[~valid["valid"]] = np.nan
    tracks = u2s(tracks, tracks_dtype)
    tracks = join_structured_arrays([tracks, valid])
    hadron_track_select = rng.choice(np.arange(5), size=(n_jets, n_tracks_per_jet))
    track_barcodes = hadrons["barcode"][np.arange(n_jets)[:, None], hadron_track_select]
    tracks["ftagTruthParentBarcode"] = track_barcodes
    tracks["ftagTruthParentBarcode"][~tracks["valid"]] = -1

    # setup electrons
    electrons_dtype = np.dtype(
        [(n, "f4") for n in electron_vars]
        + [
            ("ftagTruthOriginLabel", "i4"),
            ("ftagTruthVertexIndex", "i4"),
        ]
    )
    electrons = rng.random(shapes_electrons["inputs"])
    electrons = u2s(electrons, electrons_dtype)
    valid = rng.choice([True, False], size=shapes_electrons["valid"])
    valid = np.sort(valid, axis=-1)[:, ::-1].view(dtype=np.dtype([("valid", bool)]))
    electrons = join_structured_arrays([electrons, valid])

    # setup parameters
    params_dtype = np.dtype([(n, "f4") for n in params])
    params = rng.random(shapes_params["inputs"])
    params = u2s(params, params_dtype)
    if inc_params:
        params["mass"] = rng.choice([5, 16, 25, 40, 55], size=(n_jets))

    with h5py.File(fname, "w") as f:
        f.attrs["unique_jets"] = len(jets)
        f.attrs["config"] = "{}"
        f.create_dataset("jets", data=jets)
        if make_xbb:
            f["jets"].attrs["flavour_label"] = ["hbb", "hcc", "top", "qcd"]
        else:
            f["jets"].attrs["flavour_label"] = ["bjets", "cjets", "ujets"] + (
                ["taus"] if inc_taus else []
            )
        f.create_dataset("tracks", data=tracks)
        f.create_dataset("tracks_dr", data=tracks)
        f.create_dataset("tracks_ghost", data=tracks)
        f.create_dataset("electrons", data=electrons)
        f.create_dataset("flow", data=tracks)
        f.create_dataset("truth_hadrons", data=hadrons)
        if inc_params:
            f.create_dataset("PARAMETERS", data=params)


def as_half(typestr) -> np.dtype:
    """Cast float type to half precision."""
    t = np.dtype(typestr)
    if t.kind != "f" or t.itemsize != 2:
        return t
    return np.dtype("f2")



"""Top level training script, powered by the lightning CLI."""

import pathlib
import sys

import comet_ml  # noqa: F401
from lightning.pytorch.cli import ArgsType
from torch import save

from salt.callbacks import SaveConfigCallback
from salt.data.datamodules import SaltDataModule
from salt.modelwrapper import ModelWrapper
from salt.utils.cli import SaltCLI
from salt.utils.muP_utils.configuration_muP import get_model_path


def generateModel(args: ArgsType = None) -> None:
    config_dir = pathlib.Path(__file__).parent.parent.parent / "configs"

    if args is None:
        model_type = sys.argv.pop(1)
    else:
        model_type = args[0]
        args = args[1:]
        sys.argv = [sys.argv[0]]

    cli = SaltCLI(
        model_class=ModelWrapper,
        datamodule_class=SaltDataModule,
        save_config_callback=SaveConfigCallback,
        args=args,
        run=False,
        parser_kwargs={
            "default_env": True,
            "default_config_files": [f"{config_dir}/base.yaml"],
        },
    )
    if "temp" in model_type:
        # Store the model with the ModelWrapper
        print(f"Saving model_type={model_type} at {get_model_path(model_type)}")
        save(cli.model, get_model_path(model_type))
    else:
        # Store the model inside the ModelWrapper
        print(f"Saving model_type={model_type} at {get_model_path(model_type)}")
        save(cli.model.model, get_model_path(model_type))


if __name__ == "__main__":
    generateModel()



import torch
from torch import Tensor, nn

SYM_VARS = {"phi"}


class PositionalEncoder(nn.Module):
    def __init__(self, variables: list[str], dim: int, alpha: int = 100):
        """Positional encoder.

        Evenly share the embedding space between the different variables to be encoded.
        Any remaining dimensions are left as zeros.

        TODO: alpha should be set for each variable

        Parameters
        ----------
        variables : list[str]
            List of variables to apply the positional encoding to.
        dim : int
            Dimension of the positional encoding.
        alpha : int, optional
            Scaling factor for the positional encoding, by default 100.
        """
        super().__init__()
        self.variables = variables
        self.dim = dim
        self.alpha = alpha
        print(dim, len(self.variables))
        self.per_input_dim = self.dim // (2 * len(self.variables))
        self.last_dim = self.dim % (2 * len(self.variables))
        print(self.per_input_dim, self.last_dim)

    @torch.no_grad()
    def forward(self, inputs: Tensor):
        """Apply positional encoding to the inputs.

        Parameters
        ----------
        inputs : Tensor
            Input tensor for coordinates to be encoded.

        Returns
        -------
        Tensor
            Positional encoding of the input variables.
        """
        encodings = []
        for i, var in enumerate(self.variables):
            kwargs = {"symmetric": var in SYM_VARS}
            encodings.append(self.pos_enc(inputs[..., i], self.per_input_dim, **kwargs))

        if self.last_dim > 0:
            encodings.append(torch.zeros_like(encodings[0][..., : self.last_dim]))
        print([e.shape for e in encodings])
        return torch.cat(encodings, dim=-1)

    def pos_enc(self, xs: Tensor, dim: int, symmetric: bool = False) -> Tensor:
        """Positional encoding.

        Parameters
        ----------
        xs : Tensor
            Input tensor.
        dim : int
            Dimension of the positional encoding.
        symmetric : bool, optional
            Whether to use symmetric encoding, by default False.

        Returns
        -------
        Tensor
            Positional encoding.
        """
        xs = xs.unsqueeze(-1)
        kwargs = {"device": xs.device, "dtype": xs.dtype}
        omegas = self.alpha * torch.logspace(0, 2 / (dim) - 1, dim, 10_000, **kwargs)
        if symmetric:
            p1 = (xs.sin() * omegas).sin()
            p2 = (xs.cos() * omegas).sin()
        else:
            p1 = (xs * omegas).sin()
            p2 = (xs * omegas).cos()
        return torch.cat((p1, p2), dim=-1)



from typing import TypeAlias

from torch import BoolTensor, Tensor

Vars: TypeAlias = dict[str, list[str]]
Tensors: TypeAlias = dict[str, Tensor]
BoolTensors: TypeAlias = dict[str, BoolTensor]
NestedTensors: TypeAlias = dict[str, dict[str, Tensor]]



import argparse
from pathlib import Path

from condor_handler import CondorHandler

# Set up argument parser
parser = argparse.ArgumentParser(description="Submit batch jobs to HTCondor.")
parser.add_argument("-c", "--config", required=True, type=Path, help="Configuration file for job.")
parser.add_argument("-t", "--tag", default="salt_job", help="Tag for job to be submitted.")
parser.add_argument(
    "-e",
    "--environment",
    default="conda",
    choices=["conda", "singularity", "local"],
    help="Environment for job to be submitted.",
)
args = parser.parse_args()

# Define directories
batch_dir = Path.cwd() / "condor"
batch_path = batch_dir / "batch"
log_path = batch_dir / "batch_logs"
for directory in [batch_path, log_path]:
    directory.mkdir(parents=True, exist_ok=True)

# Set up HTCondor options
job_basedir = Path(__file__).resolve().parent.parent.parent
handler = CondorHandler(str(batch_path), str(log_path), str(job_basedir))
handler["universe"] = "vanilla"
handler["cpu"] = 1
handler["gpu"] = 2
handler["memory"] = 80_000  # 80 GiB - in MiB
handler["runtime"] = 82800  # 23h - in seconds
handler["requirements"] = 'OpSysAndVer == "CentOS7"'

# Run in singularity container?
if args.environment == "singularity":
    handler["container"] = (
        '"/cvmfs/unpacked.cern.ch/gitlab-registry.cern.ch/'
        'atlas-flavor-tagging-tools/algorithms/salt:latest"'
    )
    # check host names to determine where to mount storage
    storage_dir = Path("/tmp")  # noqa: S108
    if Path("/etc/hostname").read_text().startswith("lxplus"):
        storage_dir = Path("/eos")
    handler["containerargs"] = f'"--nv --bind {storage_dir}"'

# Construct and submit the job command
command = "cd ${BASEDIR} && export OMP_NUM_THREADS=1 && "
if args.environment == "conda":
    command += (
        "source conda/bin/activate && conda activate salt && "
        'echo "Activated environment ${CONDA_DEFAULT_ENV}" && '
    )
command += (
    'echo "CUDA_VISIBLE_DEVICES: ${CUDA_VISIBLE_DEVICES}" && '
    "cat /proc/cpuinfo | awk '/^processor/{print $3}' | tail -1 && "
    "cd ${BASEDIR}/salt && pwd && "
    f"salt fit --config {args.config.resolve()}"
)
handler.send_job(command, args.tag)



import argparse
import json

import onnx
import onnxruntime as ort


def parse_args(args):
    parser = argparse.ArgumentParser(
        description="Get metdata from a .onnx file",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        dest="file",
        type=str,
        help="Path to onnx file.",
    )
    parser.add_argument(
        "-k",
        "--key",
        type=str,
        help="Get config from the specified key.",
        default="gnn_config",
    )

    return parser.parse_args(args)


def main(args=None):
    args = parse_args(args)

    onnx_model = onnx.load(args.file)
    onnx.checker.check_model(onnx_model)

    sess_options = ort.SessionOptions()
    sess_options.graph_optimization_level = ort.GraphOptimizationLevel.ORT_DISABLE_ALL
    sess = ort.InferenceSession(args.file, sess_options, providers=ort.get_available_providers())
    meta = sess.get_modelmeta()
    name = meta.description
    config = meta.custom_metadata_map[args.key]
    info = json.loads(config)
    info = json.dumps(info, indent=1)

    print(f"Printing config for model: '{name}' and key: '{args.key}'")
    print(info)


if __name__ == "__main__":
    main()



import pytest
import torch

from salt.models import FeaturewiseTransformation


@pytest.fixture
def dense_config():
    return {"output_size": 5, "hidden_layers": [5], "activation": "ReLU"}


@pytest.fixture
def variables():
    return {"PARAMETERS": ["x"]}


def test_featurewise_forward(dense_config, variables):
    layer = "input"

    featurewise = FeaturewiseTransformation(
        layer,
        variables,
        dense_config_scale=dense_config,
        dense_config_bias=dense_config,
    )

    inputs = {"PARAMETERS": torch.randn(1, 1)}
    features = torch.rand(1, 10, 5)
    transformed_features = featurewise(inputs, features)

    expected_output_shape = features.shape
    output_shape = transformed_features.shape
    # featurewise transformations should preserve size
    assert expected_output_shape == output_shape



import scipy
import torch
import torch.nn.functional as F
from torch import Tensor, nn


@torch.jit.script
def batch_dice_cost(inputs: Tensor, targets: Tensor) -> Tensor:
    """Compute the DICE loss for the entire batch, similar to generalized IOU for masks,
    over all permutations of the inputs and targets.
    """
    inputs = inputs.sigmoid()

    # inputs has shape (B, N, C), targets has shape (B, M, C)
    # We want to compute the DICE loss for each combination of N and M for each batch
    # Using torch.einsum to handle the batched matrix multiplication
    numerator = 2 * torch.einsum("bnc,bmc->bnm", inputs, targets)

    # Compute the denominator using sum over the last dimension (C) and broadcasting
    denominator = inputs.sum(-1).unsqueeze(2) + targets.sum(-1).unsqueeze(1)

    return 1 - (numerator + 1) / (denominator + 1)


@torch.jit.script
def batch_sigmoid_ce_cost(inputs: Tensor, targets: Tensor) -> Tensor:
    """Compute the cross entropy loss over all permutations of the inputs and targets."""
    pos = F.binary_cross_entropy_with_logits(inputs, torch.ones_like(inputs), reduction="none")
    neg = F.binary_cross_entropy_with_logits(inputs, torch.zeros_like(inputs), reduction="none")
    loss = torch.einsum("bnc,bmc->bnm", pos, targets) + torch.einsum(
        "bnc,bmc->bnm", neg, (1 - targets)
    )
    return loss / inputs.shape[2]


@torch.jit.script
def batch_sigmoid_focal_cost(
    inputs: Tensor, targets: Tensor, alpha: float = -1, gamma: float = 2
) -> Tensor:
    """Compute the focal loss over all permutations of the inputs and targets for batches."""
    prob = inputs.sigmoid()
    focal_pos = ((1 - prob) ** gamma) * F.binary_cross_entropy_with_logits(
        inputs, torch.ones_like(inputs), reduction="none"
    )
    focal_neg = (prob**gamma) * F.binary_cross_entropy_with_logits(
        inputs, torch.zeros_like(inputs), reduction="none"
    )
    if alpha >= 0:
        focal_pos = focal_pos * alpha
        focal_neg = focal_neg * (1 - alpha)
    loss = torch.einsum("bnc,bmc->bnm", focal_pos, targets) + torch.einsum(
        "bnc,bmc->bnm", focal_neg, (1 - targets)
    )
    return loss / inputs.shape[2]


@torch.jit.script
def batch_mae_loss(inputs: Tensor, targets: Tensor) -> Tensor:
    """Compute the mean average error over all permutations of the inputs and targets
    for batches.
    """
    return (inputs[:, :, None] - targets[:, None, :]).abs().mean(-1)


class HungarianMatcher(nn.Module):
    """Module to compute the matching cost and solve the corresponding LSAP using the
    Hungarian algorithm.

    Based on
    - https://github.com/facebookresearch/MaskFormer
    - https://github.com/facebookresearch/Mask2Former
    """

    def __init__(
        self,
        num_classes: int,
        num_objects: int,
        loss_weights: dict,
    ):
        """Compute the optimal assignment between the targets and the predictions of the network.

        For efficiency reasons, the targets don't include the no_object null class. Because of this,
        in general, there are more predictions than targets. In this case, we do a 1-to-1 matching
        of the best predictions, while the others are un-matched (and thus treated as
        non-objects).

        Parameters
        ----------
        num_classes: int
            Number of object classes, omitting the special no_object class
        num_objects: int
            Number of object classes + 1 (the no_object class)
        loss_weights: dict
            Dictionary containing the weights for the different losses
        """
        super().__init__()
        self.num_classes = num_classes
        self.num_objects = num_objects
        self.loss_weights = loss_weights
        assert sum(self.loss_weights.values()) != 0, "Sum of loss weights must be positive"

        self.global_step = 0

    def get_batch_cost(self, preds, targets):
        """Gets the cost matrix for the whole batch.

        Parameters
        ----------
        preds : dict of tensors
            The predictions from the model
        targets : dict of tensors
            The targets for the model

        Returns
        -------
        C : tensor
            The cost matrix for the whole batch, with a large value set in the input pad region
        batch_obj_lengths : tensor
            The number of inputs in each batch element
        """
        # get some useful things
        bs = len(targets["object_class"])
        dev = preds["class_probs"].device

        obj_class_tgt = targets["object_class"].detach()
        obj_class_pred = preds["class_probs"].detach()
        mask_pred = preds["masks"].detach()
        mask_tgt = targets["masks"].detach().to(mask_pred.dtype)

        valid_obj_idx = obj_class_tgt != self.num_classes
        batch_obj_lengths = torch.sum(valid_obj_idx, dim=1)

        # compute the object class loss
        obj_class_tgt = (
            obj_class_tgt[:, : self.num_classes].unsqueeze(1).expand(-1, obj_class_pred.size(1), -1)
        )
        valid_obj_mask = obj_class_tgt != self.num_classes
        output = torch.gather(obj_class_pred, 2, obj_class_tgt * valid_obj_mask) * valid_obj_mask
        obj_class_cost = torch.zeros((bs, self.num_objects, self.num_objects), device=dev)
        obj_class_cost[:, :, : self.num_classes] = -output

        # initialize the cost matrix with the object class loss
        C = self.loss_weights["object_class_ce"] * obj_class_cost

        # add mask costs
        if self.loss_weights.get("mask_dice"):
            cost_mask_dice = batch_dice_cost(mask_pred, mask_tgt)
            C += self.loss_weights["mask_dice"] * cost_mask_dice
        if self.loss_weights.get("mask_ce"):
            cost_mask_ce = batch_sigmoid_ce_cost(mask_pred, mask_tgt)
            C += self.loss_weights["mask_ce"] * cost_mask_ce
        if self.loss_weights.get("mask_focal"):
            cost_mask_focal = batch_sigmoid_focal_cost(mask_pred, mask_tgt)
            C += self.loss_weights["mask_focal"] * cost_mask_focal

        # add regression costs
        if "regression" in preds and self.loss_weights.get("regression"):
            reg_pred = preds["regression"]
            reg_tgt = targets["regression"] * valid_obj_idx.unsqueeze(-1)
            C += self.loss_weights["regression"] * batch_mae_loss(reg_pred, reg_tgt)

        # set entries corresponding to invalid objects to nan
        # (these are removed later when running LSAP)
        batch_obj_lengths = batch_obj_lengths.unsqueeze(-1)
        col_indices = torch.arange(obj_class_cost.size(-1), device=dev).unsqueeze(0)
        null_obj_cost_mask = (col_indices < batch_obj_lengths).unsqueeze(1).expand_as(C)
        C[~null_obj_cost_mask] = torch.nan

        return C, batch_obj_lengths

    @torch.no_grad()
    def forward(self, preds, targets):
        """Compute the optimal assignment between the targets and the predictions of the network."""
        batch_size = preds["class_logits"].shape[0]

        idxs = []
        self.default_idx = set(range(self.num_objects))

        # Get the full cost matrix, then run lsap on each batch element
        full_cost, batch_N = self.get_batch_cost(preds, targets)
        full_cost = full_cost.to(torch.float32).cpu().numpy()

        for batch_idx in range(batch_size):
            # get the cost matrix for this batch element
            C = full_cost[batch_idx][:, : batch_N[batch_idx]]

            # get the optimal assignment
            idx = self.lap(C)

            idxs.append(idx)

        # get the device so we can put the indices on the same device as the predictions
        d = preds["class_logits"].device
        # format indices to allow simple indexing
        idxs = torch.tensor(idxs).to(d)
        idxs = (torch.arange(len(idxs)).unsqueeze(1).to(d), idxs)
        self.global_step += 1
        return idxs

    def lap(self, cost):
        src_idx, tgt_idx = scipy.optimize.linear_sum_assignment(cost)
        idx = src_idx[tgt_idx]
        return list(idx) + sorted(self.default_idx - set(idx))



"""Top level training script, powered by the lightning CLI."""

import pathlib

import comet_ml  # noqa: F401
from lightning.pytorch.cli import ArgsType

from salt.callbacks import SaveConfigCallback
from salt.data.datamodules import SaltDataModule
from salt.modelwrapper import ModelWrapper
from salt.utils.cli import SaltCLI


def main(args: ArgsType = None) -> None:
    config_dir = pathlib.Path(__file__).parent / "configs"

    SaltCLI(
        model_class=ModelWrapper,
        datamodule_class=SaltDataModule,
        save_config_callback=SaveConfigCallback,
        args=args,
        parser_kwargs={
            "default_env": True,
            "fit": {"default_config_files": [f"{config_dir}/base.yaml"]},
            "test": {"default_config_files": [f"{config_dir}/base.yaml"]},
        },
    )


if __name__ == "__main__":
    main()



from pathlib import Path

import torch
import yaml
from torch import nn

from salt.stypes import Tensors, Vars


class InputNorm(nn.Module):
    def __init__(
        self, norm_dict: Path, variables: Vars, global_object: str, input_map: dict[str, str]
    ) -> None:
        """Normalise inputs on the fly using a pre-computed normalisation dictionary.

        Parameters
        ----------
        norm_dict : Path
            Path to file containing normalisation parameters
        variables : dict
            Input variables for each type of input
        global_object : str
            Name of the global input object, as opposed to the constituent-level
            inputs
        input_map : dict
            Map names to the corresponding dataset names in the input h5 file.
            Set automatically by the framework.
        """
        super().__init__()
        self.variables = variables
        self.global_object = global_object
        self.NO_NORM = ["EDGE", "PARAMETERS"]
        with open(norm_dict) as f:
            self.norm_dict = yaml.safe_load(f)

        # get the keys that need to be normalised
        if input_map is None:
            input_map = {k: k for k in variables}
        keys = {input_map[k] for k in set(variables.keys())}
        keys.discard("EDGE")
        keys.discard("PARAMETERS")
        if "GLOBAL" in keys:
            keys.remove("GLOBAL")
            keys.add(self.global_object)

        # check we have all required keys in the normalisation dictionary
        if missing := keys - set(self.norm_dict):
            raise ValueError(
                f"Missing input types {missing} in {norm_dict}. Choose from"
                f" {self.norm_dict.keys()}."
            )

        # check we have all required variables for each input type
        for k, vs in variables.items():
            if k in self.NO_NORM:
                continue
            name = input_map[k]
            if k == "GLOBAL":
                name = self.global_object
            if missing := set(vs) - set(self.norm_dict[name]):
                raise ValueError(
                    f"Missing variables {missing} for {name} in {norm_dict}. Choose from"
                    f" {self.norm_dict[name].keys()}."
                )

            # store normalisation parameters with the model
            means = torch.as_tensor([self.norm_dict[name][v]["mean"] for v in vs])
            stds = torch.as_tensor([self.norm_dict[name][v]["std"] for v in vs])
            self.register_buffer(f"{k}_means", means)
            self.register_buffer(f"{k}_stds", stds)

            # check normalisation parameters are ok
            if not torch.isfinite(means).all() or not torch.isfinite(stds).all():
                raise ValueError(f"Non-finite normalisation parameters for {name} in {norm_dict}.")
            if any(stds == 0):
                raise ValueError(f"Zero standard deviation for {name} in {norm_dict}.")

    def forward(self, inputs: Tensors) -> Tensors:
        for k, x in inputs.items():
            if k in self.NO_NORM:
                continue
            inputs[k] = (x - getattr(self, f"{k}_means")) / getattr(self, f"{k}_stds")
        return inputs



import os
import shutil
from multiprocessing import Pool
from pathlib import Path

import yaml
from tqdm import tqdm


def get_temp_path(move_files_temp: str, in_path: str):
    """Create full temp file path."""
    return Path(Path(move_files_temp) / Path(in_path).name)


def copy_file(in_path: Path, out_path: Path):
    """Copy a file unless it already exists."""
    if in_path == out_path or out_path.is_file():
        return
    out_path.parent.mkdir(parents=True, exist_ok=True)
    shutil.copyfile(in_path, out_path)


def remove_file(path: Path):
    """Remove a file if it exists."""
    if path.is_file():
        path.unlink()
    else:
        print(f"No file to delete at {path}")


def remove_files_temp(train_temp_path, val_temp_path):
    remove_file(train_temp_path)
    remove_file(val_temp_path)
    val_temp_path.parent.rmdir()


def move_files_temp(move_files_temp: str, train_path: str, val_path: str):
    """Move training files to a temporary location before training.

    Set move_files_temp to "/dev/shm/your/path" for RAM.
    """
    temp_train_path = get_temp_path(move_files_temp, train_path)
    temp_val_path = get_temp_path(move_files_temp, val_path)

    copy_file(Path(train_path), temp_train_path)
    copy_file(Path(val_path), temp_val_path)


def convert_path_to_S3url(path: Path | str):
    path = str(path)
    s3_start = "s3://"
    s3_inc_start = "s3:/"

    # If there's something before the "s3:/", remove it
    if s3_inc_start in path and s3_inc_start != path[: len(s3_inc_start)]:
        assert (
            len(path.split(s3_inc_start)) <= 2
        ), "path is invalid: do not set 's3:/' in the paths!"
        path = path.split(s3_inc_start)[-1]

    if s3_start == path[: len(s3_start)]:
        return path
    if s3_inc_start == path[: len(s3_inc_start)]:
        return s3_start + path[len(s3_inc_start) :]
    return s3_start + path


def download_S3(session, bucket, file_to_load, store_path, count):
    file_to_load = file_to_load[1:] if file_to_load[0] == "/" else file_to_load
    meta_data = session.head_object(Bucket=bucket, Key=file_to_load)
    total_length = int(meta_data.get("ContentLength", 0))
    with tqdm(
        total=total_length,
        desc=f"Downloading s3://{bucket}/{file_to_load}",
        position=count,
        unit="B",
        unit_scale=True,
        unit_divisor=1024,
    ) as t:
        session.download_file(bucket, file_to_load, str(store_path), Callback=t.update)


def download_script_S3(bucket, local_path, key, file, count):
    import boto3

    target_path = Path(local_path, file.split("/")[-1])
    if not target_path.is_file():
        session = boto3.client("s3")
        download_S3(session, bucket, file, target_path, count)
    else:
        print(f'- "{file}" found locally and not downloaded.')
    return key, str(target_path)


def import_data_S3(config_path):
    with open(config_path) as file:
        cfg = yaml.safe_load(file)

    config_S3 = cfg["data"]["config_S3"]
    os.environ["AWS_ACCESS_KEY_ID"] = config_S3["pubKey"]
    os.environ["AWS_SECRET_ACCESS_KEY"] = config_S3["secKey"]
    os.environ["AWS_ENDPOINT_URL"] = config_S3["url"]

    if config_S3.get("download_S3"):
        local_path = Path(config_S3["download_path"])
        local_path.mkdir(parents=True, exist_ok=True)
        print("-" * 100)
        print(f"S3 download in progress at local path: {local_path}")
        args = [
            (config_S3["bucket"], local_path, key, cfg["data"][key], count)
            for count, key in enumerate(config_S3["download_files"])
        ]
        with Pool() as pool:
            output = pool.starmap(download_script_S3, args)
        for file, result in output:
            print(f'Downloaded {file} as {result.split("/")[-1]} at local path')
            cfg["data"][file] = str(result)

        local_config = Path(local_path, "local_base.yaml")
        with open(local_config, "w") as file:
            yaml.dump(cfg, file, sort_keys=False)
        print("Stored a local version of the config.")
        print("-" * 100, "\n")
    else:
        local_config = config_path

    return str(local_config)


def setup_S3_CLI(sc_data):
    """Setting up salt to use S3."""
    config_S3 = sc_data["config_S3"]
    os.environ["AWS_ACCESS_KEY_ID"] = config_S3["pubKey"]
    os.environ["AWS_SECRET_ACCESS_KEY"] = config_S3["secKey"]
    os.environ["AWS_ENDPOINT_URL"] = config_S3["url"]

    if config_S3.get("download_S3"):
        local_path = Path(config_S3["download_path"])
        local_path.mkdir(parents=True, exist_ok=True)
        print("-" * 100)
        print(f"S3 download in progress at local path: {local_path}")

        # Parallelise the download
        args = [
            (config_S3["bucket"], local_path, key, sc_data[key], count)
            for count, key in enumerate(config_S3["download_files"])
        ]
        with Pool() as pool:
            output = pool.starmap(download_script_S3, args)

        # Update the config
        for file, result in output:
            print(f'Downloaded {file} as {result.split("/")[-1]} at local path')
            sc_data[file] = str(result)

        print("Data part of the config updated to track the downloaded files.")
        print("-" * 100, "\n")
    return sc_data


def require_S3(path):
    with open(path) as file:
        print("Doign this")
        cfg = yaml.safe_load(file)
        if (
            "config_S3" in cfg["data"]
            and "use_S3" in cfg["data"]["config_S3"]
            and cfg["data"]["config_S3"]["use_S3"]
        ):
            return True
        return False


def require_S3_CLI(config_S3):
    """Checking whether salt requires s3."""
    if config_S3 is None:
        return False
    if config_S3.get("use_S3"):
        return True
    if config_S3.get("download_S3"):
        return True
    return False


def download_from_S3() -> None:
    config_dir = Path(__file__).parent.parent / "configs"
    config = f"{config_dir}/base.yaml"
    config = import_data_S3(config)



import pytest
import torch

from salt.models.posenc import PositionalEncoder


@pytest.fixture
def dim():
    return 32


@pytest.mark.parametrize("variables", [["x"], ["x", "y"], ["x", "y", "z"]])
def test_positional_encoder_initialization(variables):
    dim = 32
    encoder = PositionalEncoder(variables, dim)
    assert encoder.variables == variables
    assert encoder.dim == dim
    assert encoder.per_input_dim == dim // (2 * len(variables))
    assert encoder.last_dim == dim % (2 * len(variables))


@pytest.mark.parametrize("variables", [["x"], ["x", "y"], ["x", "y", "z"]])
def test_positional_encoder_call(variables, dim):
    encoder = PositionalEncoder(variables, dim)
    inputs = torch.rand(10, len(variables))
    encoded = encoder(inputs)

    # Verify output shape
    expected_shape = (10, dim)
    assert encoded.shape == expected_shape

    if dim % (2 * len(variables)) != 0:
        assert torch.all(encoded[..., -1] == 0)


@pytest.mark.parametrize("variables", [["x"], ["x", "y"], ["x", "y", "z"]])
def test_pos_enc_method(variables, dim):
    encoder = PositionalEncoder(variables, dim)
    xs = torch.rand(10)
    dim_per_var = dim // (2 * len(variables))

    # Test non-symmetric encoding
    encoded_non_sym = encoder.pos_enc(xs, dim_per_var)
    assert encoded_non_sym.shape == (10, 2 * dim_per_var)

    # Test symmetric encoding
    encoded_sym = encoder.pos_enc(xs, dim_per_var, symmetric=True)
    assert encoded_sym.shape == (10, 2 * dim_per_var)



from salt.callbacks.checkpoint import Checkpoint
from salt.callbacks.gradient_logger import GradientLoggerCallback
from salt.callbacks.maskformer_metrics import MaskformerMetrics
from salt.callbacks.performancewriter import PerformanceWriter
from salt.callbacks.predictionwriter import PredictionWriter
from salt.callbacks.saveconfig import SaveConfigCallback
from salt.callbacks.weight_logger import WeightLoggerCallback

__all__ = [
    "Checkpoint",
    "GradientLoggerCallback",
    "MaskformerMetrics",
    "PerformanceWriter",
    "PredictionWriter",
    "SaveConfigCallback",
    "WeightLoggerCallback",
]



import tempfile
from pathlib import Path

import pytest
import torch
from ftag import get_mock_file

from salt.utils import clean_logs, compare_models, repair_ckpt
from salt.utils.inputs import inputs_concat
from salt.utils.scalers import RegressionTargetScaler


def test_compare_models():
    fname_A = get_mock_file()[0]
    fname_B = get_mock_file()[0]
    args = ["--file_A", fname_A, "--file_B", fname_B, "--tagger_A", "MockTagger"]
    compare_models.main(args)


# Fixture to create a temporary directory with a specified subdirectory
@pytest.fixture
def temp_directory_with_subdir(tmp_path):
    dir1 = tmp_path / "dir1"
    dir2 = tmp_path / "dir2"
    dir3 = tmp_path / "dir3"

    dir1.mkdir()
    dir2.mkdir()
    dir3.mkdir()

    subdir_path = dir2 / "specified_subdirectory"
    subdir_path.mkdir()

    return tmp_path


# Test delete_dirs_without_subdir function
def test_delete_dirs_without_subdir(temp_directory_with_subdir):
    # Only dir1 and dir3 should be kept after calling the function
    clean_logs.delete_dirs_without_subdir(temp_directory_with_subdir, "specified_subdirectory")

    # Check if the correct directories were deleted
    assert not (temp_directory_with_subdir / "dir1").exists()
    assert (temp_directory_with_subdir / "dir2").exists()
    assert not (temp_directory_with_subdir / "dir3").exists()


# Test the main function by capturing stdout
def test_clean_logs_main(temp_directory_with_subdir):
    print(temp_directory_with_subdir)
    args = [
        "--folder_path",
        str(temp_directory_with_subdir),
        "--subdirectory",
        "specified_subdirectory",
    ]
    clean_logs.main(args=args)


def test_scaler():
    dummy_scales = {
        "var1": {"op": "log", "x_scale": 2, "x_off": 1, "op_scale": 3, "op_off": 1},
        "var2": {"op": "exp", "x_scale": 2, "x_off": 1, "op_scale": 3, "op_off": 1},
        "var3": {"op": "linear", "x_scale": 2, "x_off": 1, "op_scale": 3, "op_off": 1},
    }
    scaler = RegressionTargetScaler(dummy_scales)
    values = torch.rand(500)
    for var in dummy_scales:
        scaled = scaler.scale(var, values)
        unscaled = scaler.inverse(var, scaled)
        assert torch.allclose(values, unscaled, atol=1e-7)


# test inputs_concat function
@pytest.mark.parametrize(
    ("n_batch", "n_track", "n_jet_feat", "n_track_feat"),
    [
        (10, 20, 5, 3),  # Test case 1
        (500, 15, 10, 4),  # Test case 2
    ],
)
def test_inputs_concat(n_batch, n_track, n_jet_feat, n_track_feat):
    # Call the inputs_concat function with the specified inputs
    inputs, mask = inputs_concat(n_batch, n_track, n_jet_feat, n_track_feat)
    assert inputs.shape == (n_batch, 40, n_jet_feat + n_track_feat)
    assert mask.shape == (n_batch, n_track)
    # ensuring mask created with at least on valid tracks
    assert not torch.any(mask[:, 0])


def test_repair_checkpoint(capsys):
    with tempfile.TemporaryDirectory() as tmpdirname:
        # Setup: Create a fake checkpoint file
        ckpt_path = Path(tmpdirname) / "test.ckpt"
        state_dict = {
            "_orig_mod.layer.weight": torch.randn(5, 5),
            "_orig_mod.layer.bias": torch.randn(5),
            "layer.activation": torch.randn(5),  # Entry without "_orig_mod"
        }
        torch.save({"state_dict": state_dict}, ckpt_path)

        # test the repair_checkpoint function
        repair_ckpt.repair_checkpoint(ckpt_path)
        repaired_ckpt = torch.load(ckpt_path)
        repaired_state_dict = repaired_ckpt["state_dict"]
        for key in repaired_state_dict:
            assert not key.startswith("_orig_mod."), "Found unmodified key in state_dict"
            assert "layer.activation" in repaired_state_dict, "Unmodified key was wrongly altered"
        assert Path(str(ckpt_path) + ".bak").exists(), "Backup file not found"

        # test the main function
        repair_ckpt.main([str(ckpt_path)])
        captured = capsys.readouterr()
        output = captured.out
        assert (
            "Repaired" in output or "No need to repair" in output
        ), "Unexpected output from main function"
        if "Repaired" in output:
            assert (
                "_orig_mod.layer.weight  ==>  layer.weight" in output
            ), "Expected key rename message missing"
        else:
            assert "No need to repair" in output, "Expected 'No need to repair' message missing"



import math
from collections.abc import Mapping

import torch
from lightning.pytorch.cli import instantiate_class
from torch import BoolTensor, Size, Tensor, nn

from salt.utils.tensor_utils import masked_softmax


def merge_masks(
    q_mask: BoolTensor | None,
    kv_mask: BoolTensor | None,
    attn_mask: BoolTensor | None,
    q_shape: Size,
    k_shape: Size,
    device: torch.device,
) -> BoolTensor:
    """Create a full attention mask which incoporates the padding information.

    Using pytorch transformer convention:
        False: Real node
        True:  Zero padded
    """
    # Create the full mask which combines the attention and padding masks
    merged_mask = None

    # If either pad mask exists, create
    if q_mask is not None or kv_mask is not None:
        if q_mask is None:
            q_mask = torch.full(q_shape[:-1], False, device=device)
        if kv_mask is None:
            kv_mask = torch.full(k_shape[:-1], False, device=device)
        merged_mask = q_mask.unsqueeze(-1) | kv_mask.unsqueeze(-2)

    # If attention mask exists then it must be included
    if attn_mask is not None:
        merged_mask = attn_mask if merged_mask is None else attn_mask | merged_mask

    return merged_mask


class MultiheadAttention(nn.Module):
    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        attention: nn.Module | Mapping,
        edge_embed_dim: int = 0,
        k_dim: int | None = None,
        v_dim: int | None = None,
        out_proj: bool = True,
        update_edges: bool = False,
        muP: bool = False,
    ) -> None:
        """Generic multihead attention.

        Takes in three sequences with dim: (batch, sqeuence, features)
        - q: The primary sequence queries (determines output sequence length)
        - k: The attending sequence keys (determines incoming information)
        - v: The attending sequence values

        In a message passing sense you can think of q as your receiver nodes, v and k
        are the information coming from the sender nodes.

        When q == k(and v) this is a SELF attention operation
        When q != k(and v) this is a CROSS attention operation

        Block operations:

        1) Uses three linear layers to embed the sequences.
        - q = q_linear * q
        - k = k_linear * k
        - v = v_linear * v

        2) Outputs are reshaped to add a head dimension, and transposed for matmul.
        - dim becomes: batch, heads, sequence, features

        3) Passes these through to the attention module (message passing)
        - In standard transformers this is the scaled dot product attention
        - Also takes additional dropout layer to mask the attention

        4) Flatten out the head dimension, and optionally one more linear layer
        - results are same as if attention was done seperately for each head and concat
        - dim: batch, sequence, features*heads

        Parameters
        ----------
        embed_dim : int
            Model embedding dimension (query dim only if k_dim and v_dim also provided).
        num_heads : int
            Number of attention heads. The embed_dim is split into num_heads chunks.
        attention : nn.Module
            Type of attention (pooling operation) to use.
        edge_embed_dim: int, optional
            Model embedding dimension for edge features.
        k_dim : int, optional
            Key dimension, by default None where it assumes embed_dim
        v_dim : int, optional
            Value dimension, by default None where it assumes embed_dim
        out_proj : bool
            An optional output linear layer
        update_edges : bool, optional
            Indicate whether to update edge features.
        muP: bool, optional,
            Whether to use the muP parametrisation.
            Impacts init and scale of dot product sqrt(head_dim) -> head_dim.
            Ref: https://arxiv.org/abs/2203.03466
        """
        super().__init__()

        # Check that the dimension of each heads makes internal sense
        if embed_dim % num_heads != 0:
            raise ValueError(f"embed_dim {embed_dim} must be divisible by num_heads {num_heads}")
        if edge_embed_dim % num_heads != 0:
            raise ValueError(
                f"edge_embed_dim {edge_embed_dim} must be divisible by num_heads {num_heads}"
            )

        # Model base attributes
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.head_dim = embed_dim // num_heads
        self.out_proj = out_proj
        self.edge_embed_dim = edge_embed_dim
        self.edge_head_dim = edge_embed_dim // num_heads
        self.k_dim = k_dim or embed_dim
        self.v_dim = v_dim or embed_dim
        self.scale = self.head_dim if muP else math.sqrt(self.head_dim)
        self.update_edges = update_edges
        self.muP = muP

        # Explicitly instantiate the attention class if passed as a dictionary
        if isinstance(attention, Mapping):
            attention = instantiate_class((), attention)
        self.attention = attention

        # The different linear projection layers, output is optional
        self.linear_q = nn.Linear(self.embed_dim, self.embed_dim)
        self.linear_k = nn.Linear(self.k_dim, self.embed_dim)
        self.linear_v = nn.Linear(self.v_dim, self.embed_dim)
        if self.edge_embed_dim > 0:
            self.linear_e = nn.Linear(self.edge_embed_dim, self.num_heads)
            self.linear_g = nn.Linear(self.edge_embed_dim, self.num_heads)
            if self.update_edges:
                self.linear_e_out = nn.Linear(self.num_heads, self.edge_embed_dim)
            else:
                self.register_buffer("linear_e_out", None)
        if self.out_proj:
            self.linear_out = nn.Linear(self.embed_dim, self.embed_dim)
        else:
            self.register_buffer("linear_out", None)

        if self.muP:
            self._reset_parameters()

    def _reset_parameters(self):
        """Initialise the weights and biases for muP."""
        nn.init.constant_(self.linear_q.weight, 0)  # zero initialisation of query weights
        nn.init.normal_(self.linear_k.weight, std=(1.0 / self.k_dim) ** 0.5)
        nn.init.normal_(self.linear_v.weight, std=(1.0 / self.v_dim) ** 0.5)
        linear_list = [self.linear_q, self.linear_k, self.linear_v]
        if self.edge_embed_dim > 0:
            nn.init.normal_(self.linear_e.weight, std=(1.0 / self.edge_embed_dim) ** 0.5)
            nn.init.normal_(self.linear_g.weight, std=(1.0 / self.edge_embed_dim) ** 0.5)
            linear_list.extend([self.linear_e, self.linear_g])
            if self.update_edges:
                nn.init.normal_(self.linear_e_out.weight, std=(1.0 / self.num_heads) ** 0.5)
                linear_list.append(self.linear_e_out)
        if self.out_proj:
            nn.init.normal_(self.linear_out.weight, std=(1.0 / self.embed_dim) ** 0.5)
            linear_list.append(self.linear_out)
        for linear in linear_list:
            nn.init.constant_(linear.bias, 0.0)

    def input_projections(self, q, k, v) -> tuple:
        """Perform input linear projections, output shapes are (B,L,H,HD)."""
        shape = (k.shape[0], -1, self.num_heads, self.head_dim)
        q_proj = self.linear_q(q).view(shape).transpose(1, 2)
        k_proj = self.linear_k(k).view(shape).transpose(1, 2)
        v_proj = self.linear_v(v).view(shape).transpose(1, 2)
        return q_proj, k_proj, v_proj

    def forward(
        self,
        q: Tensor,
        k: Tensor | None = None,
        v: Tensor | None = None,
        edges: Tensor | None = None,
        q_mask: BoolTensor | None = None,
        kv_mask: BoolTensor | None = None,
        attn_mask: BoolTensor | None = None,
        attn_bias: Tensor | None = None,
    ) -> Tensor:
        """Full forward pass through the model.

        Parameters
        ----------
        q : Tensor
            The main sequence used to generate the queries.
        k : Optional[Tensor], optional
            Seperate sequence from which to generate the keys, by default None
        v : Optional[Tensor], optional
            Seperate sequence from which to generate the values, by default None
        edges : Optional[Tensor], optional
            Main sequence for edge features (used to calculate E and G)
        q_mask : Optional[BoolTensor], optional
            Shows which elements of q are real verses padded, by default None
        kv_mask : Optional[BoolTensor], optional
            Shows which elements of k and v are real verses padded, by default None
        attn_mask : Optional[BoolTensor], optional
            Extra mask for the attention (adjacency) matrix, by default None
        attn_bias : Optional[Tensor], optional
            Extra values to further augment the attention matrix, by default None

        Returns
        -------
        Tensor
            Output with the same shape as q
        """
        # If only q and q_mask are provided then we automatically apply self attention
        if k is None:
            k = q
            # warning: this isn't necessary or nan-safe in general
            # however fine how we implemented it since we use a custom masked softmax
            if kv_mask is None:
                kv_mask = q_mask
        v = v if v is not None else k

        # input shape
        b_size, _seq_len, _features = q.shape

        # Work out the masking situation, with padding, peaking, etc
        attn_mask = merge_masks(q_mask, kv_mask, attn_mask, q.shape, k.shape, q.device)

        # Apply the input projections (B,H,L,HD)
        q_proj, k_proj, v_proj = self.input_projections(q, k, v)

        # Calculate edge feature matrices E, G and reshape them to (B,L,L,H)
        if edges is not None:
            e = self.linear_e(edges)
            g = nn.functional.sigmoid(self.linear_g(edges))
            attn_bias = e if attn_bias is None else attn_bias + e

        # Calculate attention scores (B,H,Lq,Lk)
        attn_weights = self.attention(
            q_proj, k_proj, self.scale, attn_mask, attn_bias, self.update_edges
        )
        if self.update_edges:
            attn_weights, attn_scores = attn_weights

        # Apply gating to attention scores
        if edges is not None:
            attn_weights = attn_weights * g.permute(0, 3, 1, 2)

        # Use the scores for pooling and reshape (B, Lv, F)
        out = torch.matmul(attn_weights, v_proj)
        out = out.transpose(1, 2).contiguous().view(b_size, -1, self.embed_dim)

        # update edges with dot product attention scores (if desired)
        edge_out = None
        if self.update_edges:
            edge_out = self.linear_e_out(attn_scores.permute(0, 2, 3, 1))

        # Optional output layer
        if self.out_proj:
            out = self.linear_out(out)

        if edges is not None:
            return out, edge_out

        return out


class ScaledDotProductAttention(nn.Module):
    """Scaled dot product attention, commonly used in transformers.

    Contains dropout layer to stochastically mask messages, used a lot in language
    processing.

    Allows for addition of extra bias term in the attention matrix as used in the
    particle transformer: https://arxiv.org/abs/2202.03772
    """

    def __init__(self, dropout: float = 0.0) -> None:
        super().__init__()
        self.dropout = nn.Dropout(dropout)

    def forward(
        self,
        q: Tensor,
        k: Tensor,
        scale: Tensor,
        mask: BoolTensor | None = None,
        attn_bias: Tensor | None = None,
        return_scores: bool = False,
    ) -> Tensor:
        # inputs are of shape (batch, heads, sequence, head_dim)

        # dot product between queries and keys
        scores = torch.matmul(q, k.transpose(-2, -1)) / scale

        # add the bias terms if present
        if attn_bias is not None:  # Move the head dimension to the first
            scores = scores + attn_bias.permute(0, 3, 1, 2)

        # apply the dropout to the scores
        scores = self.dropout(scores)

        # softmax, use the mask to not included padded information
        attention_weights = masked_softmax(scores, mask)

        if return_scores:
            return attention_weights, scores

        return attention_weights


class GATv2Attention(nn.Module):
    """GATv2 attention, used in the original implementation of GN1.

    https://arxiv.org/abs/2105.14491
    """

    def __init__(self, num_heads: int, head_dim: int, activation: str = "SiLU") -> None:
        super().__init__()
        self.attention = nn.Parameter(torch.FloatTensor(size=(1, num_heads, 1, 1, head_dim)))
        self.activation = getattr(nn, activation)()
        nn.init.xavier_uniform_(self.attention)

    def forward(
        self,
        q: Tensor,
        k: Tensor,
        scale: Tensor,
        mask: BoolTensor | None = None,
        attn_bias: Tensor | None = None,
        return_scores: bool = False,
    ) -> Tensor:
        _ = scale
        # inputs are (B, H, Lq/k, D)
        # B, H, Lq, D = q.shape

        # sum each pair of tracks within a batch
        # shape: (B, H, Lq, Lk, D)
        summed = q.unsqueeze(-2) + k.unsqueeze(-3)

        # after activation, dot product with learned vector
        # shape: (B, H, Lq, Lk)
        scores = (self.activation(summed) * self.attention).sum(dim=-1)

        # add the optional bias
        if attn_bias is not None:
            scores = scores + attn_bias

        # softmax
        attention_weights = masked_softmax(scores, mask)

        if return_scores:
            return attention_weights, scores

        return attention_weights



"""Compare the predictions of two taggers.

You can compare two models in the same file, or between two files.

Example usage:
    - export a model using to_onnx
    - dump using the tdd (use the -p flag)
    - evaluate the model using salt on the dumped file
    - compare outputs from the tdd and salt evaluation

For more info about model validation, see the docs:
https://ftag-salt.docs.cern.ch/export/#athena-validation

"""

import argparse
from pathlib import Path

import numpy as np
from ftag import Cuts
from ftag.hdf5 import H5Reader


def parse_args(args=None):
    parser = argparse.ArgumentParser(
        description="Script to compare predictions between two models.",
    )

    parser.add_argument(
        "--file_A", required=True, type=Path, help="Path to first evaluation h5 file."
    )
    parser.add_argument(
        "--file_B",
        required=False,
        type=Path,
        help="Path to second evaluation h5 file (defaults to file_A if not provided).",
    )
    parser.add_argument(
        "--tagger_A",
        required=True,
        type=str,
        help="Name of the first tagger.",
    )
    parser.add_argument(
        "--tagger_B",
        required=False,
        type=str,
        help="Name of the second tagger (defaults to tagger_A if not provided).",
    )
    parser.add_argument(
        "--gaussian_regression",
        help="Export gaussian regression model.",
        action="store_true",
    )
    parser.add_argument(
        "--vars",
        default=["b", "c", "u"],
        nargs="+",
        help="Which variables to compare (defaults to pb, pc, pu).",
    )
    parser.add_argument(
        "--cuts",
        default=["n_tracks <= 40"],
        nargs="+",
        help="Which cuts to apply (defaults to 'n_tracks <= 40').",
    )
    parser.add_argument(
        "-n",
        "--num",
        default=-1,
        type=float,
        help="Compare this many jets (defaults to all).",
    )

    args = parser.parse_args(args)
    args.file_B = args.file_B or args.file_A
    args.tagger_B = args.tagger_B or args.tagger_A

    return args


def main(args=None):
    args = parse_args(args)
    if args.file_A == args.file_B and args.tagger_A == args.tagger_B:
        raise ValueError("Attempted to compare the same model!")

    if args.gaussian_regression:
        vars_A = [f"{args.tagger_A}_{v}" for v in args.vars]
        vars_B = [f"{args.tagger_B}_{v}" for v in args.vars]
    else:
        vars_A = [f"{args.tagger_A}_p{v}" for v in args.vars]
        vars_B = [f"{args.tagger_B}_p{v}" for v in args.vars]

    cuts = Cuts.from_list(args.cuts)

    # load the data
    reader_A = H5Reader(args.file_A)
    reader_B = H5Reader(args.file_B)
    jets_A = reader_A.load({"jets": vars_A}, cuts=cuts, num_jets=args.num)["jets"]
    jets_B = reader_B.load({"jets": vars_B}, cuts=cuts, num_jets=args.num)["jets"]

    diff_regions = np.array([
        [float(f"1e-{i}"), float(f"5e-{i}")] for i in range(7, 1, -1)
    ]).flatten()

    print(f'\nComparing {len(jets_A):,} jets from "{args.file_A}" and "{args.file_B}"...\n')
    for var_A, var_B in zip(vars_A, vars_B, strict=True):
        array_A = jets_A[var_A]
        array_B = jets_B[var_B]
        if array_A.dtype != np.float32:
            raise ValueError(
                f"{var_A} in {args.file_A} is not float32. Please compare at full precision."
            )
        if array_B.dtype != np.float32:
            raise ValueError(
                f"{var_B} in {args.file_B} is not float32. Please compare at full precision."
            )

        # make sure there are no NaNs
        if np.isnan(array_A).any():
            raise ValueError(f"{var_A} in {args.file_A} contains NaNs.")
        if np.isnan(array_B).any():
            raise ValueError(f"{var_B} in {args.file_B} contains NaNs.")

        print(f"Comparing {var_A} and {var_B}...")
        diff = abs(array_A - array_B)
        for diff_region in diff_regions:
            selected = diff[diff > diff_region]
            pct = len(selected) / len(diff)
            print(f"Differences of {diff_region:.1e}: {pct:.2%}")
            if pct == 0:
                break
        print(f"max {diff.max():.2e} | mean {diff.mean():.2e} | median {np.median(diff):.2e}")
        print()

        np.testing.assert_allclose(array_A, array_B, rtol=1e-4, atol=1e-6)


if __name__ == "__main__":
    main()



from torch import Tensor, nn

from salt.utils.tensor_utils import attach_context


class Dense(nn.Module):
    def __init__(
        self,
        input_size: int,
        output_size: int | None = None,
        hidden_layers: list[int] | None = None,
        hidden_dim_scale: int = 2,
        activation: str = "ReLU",
        final_activation: str | None = None,
        dropout: float = 0.0,
        bias: bool = True,
        context_size: int = 0,
        muP: bool = False,
    ) -> None:
        """A fully connected feed forward neural network, which can take
        in additional contextual information.

        Parameters
        ----------
        input_size : int
            Input size
        output_size : int
            Output size. If not specified this will be the same as the input size.
        hidden_layers : list, optional
            Number of nodes per layer, if not specified, the network will have
            a single hidden layer with size `input_size * hidden_dim_scale`.
        hidden_dim_scale : int, optional
            Scale factor for the hidden layer size.
        activation : str
            Activation function for hidden layers.
            Must be a valid torch.nn activation function.
        final_activation : str, optional
            Activation function for the output layer.
            Must be a valid torch.nn activation function.
        dropout : float, optional
            Apply dropout with the supplied probability.
        bias : bool, optional
            Whether to use bias in the linear layers.
        context_size : int
            Size of the context tensor, 0 means no context information is provided.
        muP: bool, optional,
            Whether to use the muP parametrisation (impacts initialisation).
        """
        super().__init__()

        if output_size is None:
            output_size = input_size
        if hidden_layers is None:
            hidden_layers = [input_size * hidden_dim_scale]

        # Save the networks input and output sizes
        self.input_size = input_size
        self.output_size = output_size
        self.context_size = context_size
        self.muP = muP

        # build nodelist
        self.node_list = [input_size + context_size, *hidden_layers, output_size]

        # input and hidden layers
        layers = []

        num_layers = len(self.node_list) - 1
        for i in range(num_layers):
            if dropout:
                layers.append(nn.Dropout(dropout))

            # linear projection
            layers.append(nn.Linear(self.node_list[i], self.node_list[i + 1], bias=bias))

            # activation for all but the final layer
            if i != num_layers - 1:
                layers.append(getattr(nn, activation)())

            # final layer: return logits by default, or activation if specified
            elif final_activation:
                layers.append(getattr(nn, final_activation)())

        # build the net
        self.net = nn.Sequential(*layers)

        if self.muP:
            self._reset_parameters()

    def forward(self, x: Tensor, context: Tensor | None = None) -> Tensor:
        if self.context_size:
            x = attach_context(x, context)
        return self.net(x)

    def _reset_parameters(self):
        """Initialise the weights and biases for muP."""
        for layer in self.net:
            if isinstance(layer, nn.Linear):
                std = 1.0 / layer.weight.shape[0] ** 0.5
                nn.init.normal_(layer.weight, std=std)
                nn.init.constant_(layer.bias, 0)
